import math
import os
from itertools import repeat
from joblib import delayed, Parallel

from omegaml.runtimes.loky import OmegaRuntimeBackend
from omegaml.util import PickableCollection

OmegaRuntimeBackend = OmegaRuntimeBackend  # noqa
default_chunksize = int(1e6)


def dfchunker(df, size=default_chunksize):
    """ chunk a dataframe as in iterator """
    return (df.iloc[pos:pos + size] for pos in range(0, len(df), size))


def insert_chunk(job):
    """
    insert one chunk of data

    :param job: the (dataframe, mongo_url, collection_name) tuple. mongo_url
                should include the database name, as the collection is taken
                from the default database of the connection.
    """
    # note we do not catch exceptions as we want to propagate errors back to caller
    # rationale: if one chunk insert fails, all should fail and user be notified
    sdf, collection = job
    result = collection.insert_many(sdf.to_dict(orient='records'))
    return len(result.inserted_ids)


def fast_insert(df, omstore, name, chunksize=default_chunksize):
    """
    fast insert of dataframe to mongodb

    Depending on size use single-process or multiprocessing. Typically
    multiprocessing is faster on datasets with > 10'000 data elements
    (rows x columns). Note this may max out your CPU and may use
    processor count * chunksize of additional memory. The chunksize is
    set to 10'000. The processor count is the default used by multiprocessing,
    typically the number of CPUs reported by the operating system.

    :param df: dataframe
    :param omstore: the OmegaStore to use. will be used to get the mongo_url
    :param name: the dataset name in OmegaStore to use. will be used to get the
    collection name from the omstore
    """
    # this is the fastest implementation (pool)
    # #records	pool	thread/wo copy	thread/w copy	pool w=0	pool dict	no chunking
    # 0.1m        1.47     2.06            2.17           1.59        2.11        2.28
    # 1m         17.4     19.8            20.6            16         17.8        22.2
    # 10m       149      193             183             177        213         256
    # based on
    # df = pd.DataFrame({'x': range(rows)})
    # om.datasets.put(df, 'test', replace=True) # no chunking: chunksize=False
    # - pool mp Pool, passes copy of df chunks, to_dict in pool processes
    # - thread/wo copy uses ThreadPool, shared memory on df
    # - thread/w copy uses ThreadPool, copy of chunks
    # - pool w=0 disables the mongo write concern
    # - pool dict performs to_dict on chunking, passes list of json docs pools just insert
    # - no chunking sets chunksize=False
    if chunksize and len(df) * len(df.columns) > chunksize:
        collection = PickableCollection(omstore.collection(name))
        # we crossed upper limits of single threaded processing, use a Pool
        # use the cached pool
        # use at least 2 processes for parallelism, at most half of available cores
        n_jobs = max(2, math.ceil(os.cpu_count() / 2))
        jobs = zip(dfchunker(df, size=chunksize), repeat(collection))
        approx_jobs = int(len(df) / chunksize)
        # we use multiprocessing backend because
        with Parallel(n_jobs=n_jobs, backend='omegaml', verbose=False) as p:
            runner = delayed(insert_chunk)
            p_jobs = (runner(job) for job in jobs)
            p._job_count = approx_jobs
            p(p_jobs)
    else:
        # still within bounds for single threaded inserts
        omstore.collection(name).insert_many(df.to_dict(orient='records'))


# ensure loky backend is registered
from omegaml.runtimes.loky import OmegaRuntimeBackend  # noqa

OmegaRuntimeBackend = OmegaRuntimeBackend  # noqa



import builtins
import datetime
import flask
import numpy as np
import pandas as pd
import re
from flask import Blueprint
from flask_restx import Resource, Namespace, fields, Api
from mongoengine import DoesNotExist
from urllib.parse import unquote
from werkzeug.exceptions import NotFound

from omegaml.backends.restapi.asyncrest import AsyncTaskResourceMixin, AsyncResponseMixin, resolve
from omegaml.restapi.util import OmegaResourceMixin, strict, AnyObject
from omegaml.util import isTrue

omega_bp = Blueprint('omega-api', __name__)
omega_api = api = Api(omega_bp)

PredictInput = strict(api).model('ModelInputSchema', {
    'columns': fields.List(fields.String),
    'data': fields.List(fields.Raw),
    'shape': fields.List(fields.Integer),
})

PredictOutput = api.model('PredictOutput', {
    'model': fields.String,
    'result': fields.Raw,
    'resource_uri': fields.String,
})

TaskInput = strict(api).model('TaskInput', {
    'resource_uri': fields.String,
})

TaskOutput = strict(api).model('TaskOutput', {
    'task_id': fields.String,
    'status': fields.String,
    'response': fields.Raw
})

DatasetInput = api.model('DatasetInput', {
    'data': fields.Raw,
    'dtypes': fields.Raw,
    'append': fields.Boolean,
})

DatasetIndex = api.model('DatasetIndex', {
    'values': fields.List(fields.Raw),
    'type': fields.String,
})

DatasetQueryOutput = api.model('DatasetQueryOutput', {
    'data': fields.Raw,
    'index': fields.Nested(DatasetIndex)
})

ScriptInput = api.model('ScriptInput', {
})

ScriptOutput = api.model('ScriptOutput', {
    'resource_uri': fields.String,
    'script': fields.String,
    'result': fields.Raw,
    'runtimes': fields.Float,
    'started': fields.DateTime,
    'ended': fields.DateTime,
})

ServiceInput = api.model('ServiceInput', {
})

ServiceOutput = api.model('ServiceOutput', {
    '*': AnyObject
})

JobInput = api.model('JobInput', {
})

JobOutput = api.model('JobOutput', {
    'resource_uri': fields.String,
    'job': fields.String,
    'source_job': fields.String,
    'job_results': fields.Raw,
    'job_runs': fields.List(fields.Raw),
    'created': fields.DateTime,
})


@api.route('/api/v1/ping')
class PingResource(Resource):
    def get(self):
        dt = datetime.datetime.now()
        return {'ping': dt.isoformat()}


@api.route('/api/v1/task/<string:taskid>/<string:action>')
class TaskResource(OmegaResourceMixin, AsyncTaskResourceMixin, Resource):
    @api.expect(TaskInput, validate=False)
    @api.marshal_with(TaskOutput)
    def get(self, taskid, action):
        return self.process_task_action(taskid, action, request=flask.request)

    def resolve_resource_method(self, taskid, context):
        # this maps a tasks resource_uri to a resource method to resolve the async result
        resourceUri = api.payload.get('resource_uri')
        path, res_kwargs = resolve(resourceUri)
        URI_METHOD_MAP = {
            r'/api/.*/model/.*/.*/?$': ('_generic_model_resource', 'prepare_result', {'pk': 'model_id'}),
            r'/api/.*/job/.*/run/?$': ('_generic_job_resource', 'prepare_result', {'pk': 'job_id'}),
            r'/api/.*/script/.*/run/?$': ('_generic_script_resource', 'prepare_result', {'pk': 'script_id'}),
            # services are also exposed as top-level services, see ServiceResoure
            r'/api/.*/service/.*/(.*)?$': ('_generic_service_resource', 'prepare_result', {'pk': 'resource_id'}),
            r'/api/service/.*/(.*)?$': ('_generic_service_resource', 'prepare_result', {'pk': 'resource_id'}),
        }
        for regexp, specs in URI_METHOD_MAP.items():
            if re.match(regexp.replace(r'/', r'\/'), resourceUri):
                resource_name, method_name, kwargs_map = specs
                resource_pk = kwargs_map.get('pk')
                meth = self._get_resource_method(resource_name, method_name)
                return meth, dict(resource_name=res_kwargs.get(resource_pk))
        raise NotFound('unknown resource {}'.format(resourceUri))


@api.route('/api/v1/model/<path:model_id>/<string:action>', methods=['GET', 'PUT'])
@api.route('/api/v1/model/<path:model_id>/', defaults={'action': 'metadata'}, methods=['GET'])
class ModelResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(PredictInput, validate=False)
    @api.marshal_with(PredictOutput)
    def get(self, model_id, action=None):
        return self.create_response_from_resource('_generic_model_resource', action, 'model', model_id)

    @api.expect(PredictInput, validate=False)
    @api.marshal_with(PredictOutput)
    def put(self, model_id, action=None):
        return self.create_response_from_resource('_generic_model_resource', action, 'model', model_id)


@api.route('/api/v1/script/<path:script_id>/run', methods=['POST'])
class ScriptResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(ScriptInput, validate=False)
    @api.marshal_with(ScriptOutput)
    def post(self, script_id):
        return self.create_response_from_resource('_generic_script_resource', 'run', 'script', script_id)


@api.route('/api/service/<path:resource_id>', defaults={'action': '*'}, methods=['GET', 'PUT', 'POST', 'DELETE'])
@api.route('/api/service/<path:resource_id>/', defaults={'action': '*'}, methods=['GET', 'PUT', 'POST', 'DELETE'])
@api.route('/api/service/<path:resource_id>/<string:action>', methods=['GET', 'PUT', 'POST', 'DELETE'])
@api.route('/api/v1/service/<path:resource_id>', defaults={'action': '*'}, methods=['GET', 'PUT', 'POST', 'DELETE'])
@api.route('/api/v1/service/<path:resource_id>/', defaults={'action': '*'}, methods=['GET', 'PUT', 'POST', 'DELETE'])
@api.route('/api/v1/service/<path:resource_id>/<string:action>', methods=['GET', 'PUT', 'POST', 'DELETE'])
# we expose service resources as /api/v1/service and /api/service
# rationale: this is user-defined, and /v1/ does not make sense in this case
# however in light of consistency, we also provide /v1/
class ServiceResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(ServiceInput, validate=False)
    @api.marshal_with(ServiceOutput)
    def get(self, resource_id, action):
        action = 'get' if action == '*' else action
        return self.create_response_from_resource('_generic_service_resource', action, 'service', resource_id)

    @api.expect(ServiceInput, validate=False)
    @api.marshal_with(ServiceOutput)
    def post(self, resource_id, action):
        action = 'post' if action == '*' else action
        return self.create_response_from_resource('_generic_service_resource', action, 'service', resource_id)

    @api.expect(ServiceInput, validate=False)
    @api.marshal_with(ServiceOutput)
    def put(self, resource_id, action):
        action = 'put' if action == '*' else action
        return self.create_response_from_resource('_generic_service_resource', action, 'service', resource_id)

    @api.expect(ServiceInput, validate=False)
    @api.marshal_with(ServiceOutput)
    def delete(self, resource_id, action):
        action = 'delete' if action == '*' else action
        return self.create_response_from_resource('_generic_service_resource', action, 'service', resource_id)


@api.route('/api/v1/job/<path:job_id>/run', methods=['POST'])
@api.route('/api/v1/job/<path:job_id>/', methods=['GET'])
@api.route('/api/v1/job/', methods=['GET'], defaults={'job_id': None})
class JobResource(OmegaResourceMixin, AsyncResponseMixin, Resource):
    result_uri = '/api/v1/task/{id}/result'

    @api.expect(JobInput, validate=False)
    @api.marshal_with(JobOutput)
    def get(self, job_id):
        action = 'list' if job_id is None else 'metadata'
        return self.create_response_from_resource('_generic_job_resource', action, 'job', job_id)

    @api.expect(JobInput, validate=False)
    @api.marshal_with(JobOutput)
    def post(self, job_id):
        return self.create_response_from_resource('_generic_job_resource', 'run', 'job', job_id)


@api.route('/api/v1/dataset/<path:dataset_id>')
class DatasetResource(OmegaResourceMixin, Resource):
    def _restore_filter(self, om, fltparams, name):
        """
        restore filter kwargs for query in om.datasets.get
        """
        # -- get filters as specified on request query args
        fltkwargs = {k: v for k, v in fltparams.items()
                     if k not in ['orient', 'limit', 'skip', 'page']}
        # -- get dtypes of dataframe and convert filter values
        metadata = om.datasets.metadata(name)
        kind_meta = metadata.kind_meta or {}
        dtypes = kind_meta.get('dtypes')
        # get numpy/python typemap. this is required for Py3 support
        # adopted from https://stackoverflow.com/a/34919415
        np_typemap = {v: getattr(builtins, k)
                      for k, v in np.sctypeDict.items()
                      if k in vars(builtins)}
        for k, v in fltkwargs.items():
            # -- get column name without operator (e.g. x__gt => x)
            col = k.split('__')[0]
            value = unquote(v)
            dtype = dtypes.get(str(col))
            if dtype:
                # -- get dtyped value and convert to python type
                value = np_typemap.get(getattr(np, dtype, str), str)(value)
                fltkwargs[k] = value
        return fltkwargs

    @api.marshal_with(DatasetQueryOutput)
    def get(self, dataset_id):
        om = self._omega
        orient = flask.request.args.get('orient', 'dict')
        fltkwargs = self._restore_filter(om, flask.request.args, dataset_id)
        df = om.datasets.getl(dataset_id, filter=fltkwargs).value
        # get index values as python types to support Py3
        index_values = list(df.index.astype('O').values)
        index_type = type(df.index).__name__
        # get data set values as python types to support Py3
        df = df.reset_index(drop=True)
        df.index = df.index.astype('O')
        # convert nan to None
        # https://stackoverflow.com/a/34467382
        data = df.where(pd.notnull(df), None).astype('O').to_dict(orient)
        return {
            'data': data,
            'index': {
                'values': index_values,
                'type': index_type,
            }
        }

    @api.expect(DatasetInput, validate=True)
    @api.response(200, 'updated')
    def put(self, dataset_id):
        om = self._omega
        orient = flask.request.args.get('orient', 'dict')
        if 'append' in flask.request.args:
            append = isTrue(flask.request.args.get('append', 'true'))
        else:
            append = isTrue(api.payload.get('append', 'true'))
        dtypes = api.payload.get('dtypes')
        if orient == 'dict':
            orient = 'columns'
        if dtypes:
            # due to https://github.com/pandas-dev/pandas/issues/14655#issuecomment-260736368
            dtypes = {k: np.dtype(v) for k, v in dtypes.items()}
        df = pd.DataFrame.from_dict(api.payload.get('data'),
                                    orient=orient).astype(dtypes)
        om.datasets.put(df, dataset_id, append=append)
        return '', 200

    @api.response(200, 'updated')
    @api.response(404, 'does not exist')
    def delete(self, dataset_id):
        om = self._omega
        try:
            om.datasets.drop(dataset_id)
        except DoesNotExist:
            status = 404
        else:
            status = 200
        return '', status



import os

from flask import Flask
from flask_restx import Api
from werkzeug.utils import redirect


def create_app(*args, **kwargs):
    from omegaml.restapi.resources import omega_bp

    app = Flask(__name__)
    # ensure slashes in URIs are matched as specified
    # see https://stackoverflow.com/a/33285603/890242
    app.url_map.strict_slashes = True
    # use Flask json encoder to support datetime
    app.config['RESTX_JSON'] = {'cls': app.json_encoder}
    app.register_blueprint(omega_bp)

    @app.route('/docs')
    def docs():
        return redirect("https://omegaml.github.io/omegaml/", code=302)

    return app


def serve_objects():
    from omegaml.restapi import resource_filter
    import re

    specs = os.environ.get('OMEGA_RESTAPI_FILTER')
    if specs:
        respecs = [re.compile(s) for s in specs.split(';') if s]
        resource_filter.extend(respecs)
    return create_app()



import unittest

from omegaml.defaults import update_from_obj


class MiscTests(unittest.TestCase):
    def test_update_obj_dict(self):
        # source and target are dicts, scalar values
        source = {
            'FOO': 'bar'
        }
        target = {
            'FOO': 'baz'
        }
        update_from_obj(source, target)
        self.assertEqual(target['FOO'], 'bar')
        # source is dict, target is dict, dict values replace
        source = {
            'FOO': dict(sub='bar')
        }
        target = {
            'FOO': dict(sub='baz')
        }
        update_from_obj(source, target)
        self.assertEqual(target['FOO'], dict(sub='bar'))
        # source is dict, target is dict, dict values merge
        source = {
            'FOO': dict(othersub='bar')
        }
        target = {
            'FOO': dict(sub='baz')
        }
        update_from_obj(source, target)
        self.assertEqual(target['FOO'], dict(sub='baz', othersub='bar'))
        # source is dict, target is dict, dict values delete
        source = {
            'FOO': dict(othersub='bar', sub='__delete__')
        }
        target = {
            'FOO': dict(sub='baz')
        }
        update_from_obj(source, target)
        self.assertEqual(target['FOO'], dict(othersub='bar'))

    def test_update_obj_attrs(self):
        # source is dict, target is attributes, scalar values
        source = {
            'FOO': 'bar'
        }
        target = MiscTests
        target.FOO = 'baz'
        update_from_obj(source, target)
        self.assertEqual(target.FOO, 'bar')
        # source is dict, target is obj, merge
        source = {
            'FOO': dict(othersub='bar')
        }
        target.FOO = dict(sub='baz')
        update_from_obj(source, target)
        self.assertEqual(target.FOO, dict(sub='baz', othersub='bar'))
        # source is dict, target is obj, merge delete
        source = {
            'FOO': dict(othersub='bar', sub='__delete__')
        }
        target.FOO = dict(sub='baz')
        update_from_obj(source, target)
        self.assertEqual(target.FOO, dict(othersub='bar'))



if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import

import logging
from celery import Celery
from socket import gethostname

from omegaml.util import dict_merge

logger = logging.getLogger(__name__)


class CeleryTask(object):
    """
    A thin wrapper for a Celery.Task object

    This is so that we can collect common delay arguments on the
    .task() call
    """

    def __init__(self, task, kwargs):
        """

        Args:
            task (Celery.Task): the celery task object
            kwargs (dict): optional, the kwargs to pass to apply_async
        """
        self.task = task
        self.kwargs = dict(kwargs)

    def _apply_kwargs(self, task_kwargs, celery_kwargs):
        # update task_kwargs from runtime's passed on kwargs
        # update celery_kwargs to match celery routing semantics
        task_kwargs.update(self.kwargs.get('task', {}))
        celery_kwargs.update(self.kwargs.get('routing', {}))
        if 'label' in celery_kwargs:
            celery_kwargs['queue'] = celery_kwargs['label']
            del celery_kwargs['label']

    def _apply_auth(self, args, kwargs, celery_kwargs):
        from omegaml.client.auth import AuthenticationEnv
        AuthenticationEnv.active().taskauth(args, kwargs, celery_kwargs)

    def apply_async(self, args=None, kwargs=None, **celery_kwargs):
        """

        Args:
            args (tuple): the task args
            kwargs (dict): the task kwargs
            celery_kwargs (dict): apply_async kwargs, e.g. routing

        Returns:
            AsyncResult
        """
        args = args or tuple()
        kwargs = kwargs or {}
        self._apply_kwargs(kwargs, celery_kwargs)
        self._apply_auth(args, kwargs, celery_kwargs)
        return self.task.apply_async(args=args, kwargs=kwargs, **celery_kwargs)

    def delay(self, *args, **kwargs):
        """
        submit the task with args and kwargs to pass on

        This calls task.apply_async and passes on the self.kwargs.
        """
        return self.apply_async(args=args, kwargs=kwargs)

    def signature(self, args=None, kwargs=None, immutable=False, **celery_kwargs):
        """ return the task signature with all kwargs and celery_kwargs applied
        """
        self._apply_kwargs(kwargs, celery_kwargs)
        sig = self.task.signature(args=args, kwargs=kwargs, **celery_kwargs, immutable=immutable)
        return sig

    def run(self, *args, **kwargs):
        return self.delay(*args, **kwargs)


class OmegaRuntime(object):
    """
    omegaml compute cluster gateway
    """

    def __init__(self, omega, bucket=None, defaults=None, celeryconf=None):
        from omegaml.util import settings

        self.omega = omega
        defaults = defaults or settings()
        self.bucket = bucket
        self.pure_python = getattr(defaults, 'OMEGA_FORCE_PYTHON_CLIENT', False)
        self.pure_python = self.pure_python or self._client_is_pure_python()
        # initialize celery as a runtimes
        taskpkgs = defaults.OMEGA_CELERY_IMPORTS
        celeryconf = celeryconf or defaults.OMEGA_CELERY_CONFIG
        # ensure we use current value
        celeryconf['CELERY_ALWAYS_EAGER'] = bool(defaults.OMEGA_LOCAL_RUNTIME)
        self.celeryapp = Celery('omegaml')
        self.celeryapp.config_from_object(celeryconf)
        # needed to get it to actually load the tasks
        # https://stackoverflow.com/a/35735471
        self.celeryapp.autodiscover_tasks(taskpkgs, force=True)
        self.celeryapp.finalize()
        # temporary requirements, use .require() to set
        self._require_kwargs = dict(task={}, routing={})
        # fixed default arguments, use .require(always=True) to set
        self._task_default_kwargs = dict(task={}, routing={})
        # default routing label
        self._default_label = self.celeryapp.conf.get('CELERY_DEFAULT_QUEUE')

    def __repr__(self):
        return 'OmegaRuntime({})'.format(self.omega.__repr__())

    @property
    def auth(self):
        return None

    @property
    def _common_kwargs(self):
        common = dict(self._task_default_kwargs)
        common['task'].update(pure_python=self.pure_python, __bucket=self.bucket)
        common['task'].update(self._require_kwargs['task'])
        common['routing'].update(self._require_kwargs['routing'])
        return common

    @property
    def _inspect(self):
        return self.celeryapp.control.inspect()

    @property
    def is_local(self):
        return self.celeryapp.conf['CELERY_ALWAYS_EAGER']

    def mode(self, local=None, logging=None):
        """ specify runtime modes

        Args:
            local (bool): if True, all execution will run locally, else on
               the configured remote cluster
            logging (bool|str|tuple): if True, will set the root logger output
               at INFO level; a single string is the name of the logger,
               typically a module name; a tuple (logger, level) will select
               logger and the level. Valid levels are INFO, WARNING, ERROR,
               CRITICAL, DEBUG

        Usage::

            # run all runtime tasks locally
            om.runtime.mode(local=True)

            # enable logging both in local and remote mode
            om.runtime.mode(logging=True)

            # select a specific module and level)
            om.runtime.mode(logging=('sklearn', 'DEBUG'))

            # disable logging
            om.runtime.mode(logging=False)
        """
        if local is not None:
            self.celeryapp.conf['CELERY_ALWAYS_EAGER'] = local
        self._task_default_kwargs['task']['__logging'] = logging
        return self

    def _client_is_pure_python(self):
        try:
            import pandas as pd
            import numpy as np
            import sklearn
        except Exception as e:
            logging.getLogger().info(e)
            return True
        else:
            return False

    def _sanitize_require(self, value):
        # convert value into dict(label=value)
        if isinstance(value, str):
            return dict(label=value)
        if isinstance(value, (list, tuple)):
            return dict(*value)
        return value

    def require(self, label=None, always=False, routing=None, task=None, **kwargs):
        """
        specify requirements for the task execution

        Use this to specify resource or routing requirements on the next task
        call sent to the runtime. Any requirements will be reset after the
        call has been submitted.

        Args:
            always (bool): if True requirements will persist across task calls. defaults to False
            label (str): the label required by the worker to have a runtime task dispatched to it.
               'local' is equivalent to calling self.mode(local=True).
            task (dict): if specified applied to the task kwargs
            kwargs: requirements specification that the runtime understands

        Usage:
            om.runtime.require(label='gpu').model('foo').fit(...)

        Returns:
            self
        """
        if label is not None:
            if label == 'local':
                self.mode(local=True)
            else:
                self.mode(local=False)
            routing = routing or {
                'label': label or self._default_label
            }
        if task or routing:
            task = task or {}
            routing = routing or {}
            if always:
                self._task_default_kwargs['routing'].update(routing)
                self._task_default_kwargs['task'].update(task)
            else:
                self._require_kwargs['routing'].update(routing)
                self._require_kwargs['task'].update(task)
        return self

    def model(self, modelname, require=None):
        """
        return a model for remote execution

        Args:
            modelname (str): the name of the object in om.models
            require (dict): routing requirements for this job

        Returns:
            OmegaModelProxy
        """
        from omegaml.runtimes.modelproxy import OmegaModelProxy
        self.require(**self._sanitize_require(require)) if require else None
        return OmegaModelProxy(modelname, runtime=self)

    def job(self, jobname, require=None):
        """
        return a job for remote execution

        Args:
            jobname (str): the name of the object in om.jobs
            require (dict): routing requirements for this job

        Returns:
            OmegaJobProxy
        """
        from omegaml.runtimes.jobproxy import OmegaJobProxy

        self.require(**self._sanitize_require(require)) if require else None
        return OmegaJobProxy(jobname, runtime=self)

    def script(self, scriptname, require=None):
        """
        return a script for remote execution

        Args:
            scriptname (str): the name of object in om.scripts
            require (dict): routing requirements for this job

        Returns:
            OmegaScriptProxy
        """
        from omegaml.runtimes.scriptproxy import OmegaScriptProxy

        self.require(**self._sanitize_require(require)) if require else None
        return OmegaScriptProxy(scriptname, runtime=self)

    def experiment(self, experiment, provider=None, implied_run=True):
        """ set the tracking backend and experiment

        Args:
            experiment (str): the name of the experiment
            provider (str): the name of the provider

        Returns:
            OmegaTrackingProxy
        """
        from omegaml.runtimes.trackingproxy import OmegaTrackingProxy
        # tracker implied_run means we are using the currently active run, i.e. with block will call exp.start()
        tracker = OmegaTrackingProxy(experiment, provider=provider, runtime=self, implied_run=implied_run)
        return tracker

    def task(self, name, **kwargs):
        """
        retrieve the task function from the celery instance

        Args:
            name (str): a registered celery task as ``module.tasks.task_name``
            kwargs (dict): routing keywords to CeleryTask.apply_async

        Returns:
            CeleryTask
        """
        taskfn = self.celeryapp.tasks.get(name)
        assert taskfn is not None, "cannot find task {name} in Celery runtime".format(**locals())
        kwargs = dict_merge(self._common_kwargs, dict(routing=kwargs))
        task = CeleryTask(taskfn, kwargs)
        self._require_kwargs = dict(routing={}, task={})
        return task

    def settings(self, require=None):
        """ return the runtimes's cluster settings
        """
        self.require(**require) if require else None
        return self.task('omegaml.tasks.omega_settings').delay().get()

    def ping(self, *args, require=None, wait=True, **kwargs):
        """
        ping the runtime

        Args:
            args (tuple): task args
            require (dict): routing requirements for this job
            wait (bool): if True, wait for the task to return, else return
                AsyncResult
            kwargs (dict): task kwargs

        Returns:
            * response (dict) for wait=True
            * AsyncResult for wait=False
        """
        self.require(**require) if require else None
        promise = self.task('omegaml.tasks.omega_ping').delay(*args, **kwargs)
        return promise.get() if wait else promise

    def enable_hostqueues(self):
        """ enable a worker-specific queue on every worker host

        Returns:
            list of labels (one entry for each hostname)
        """
        control = self.celeryapp.control
        inspect = control.inspect()
        active = inspect.active()
        queues = []
        for worker in active.keys():
            hostname = worker.split('@')[-1]
            control.cancel_consumer(hostname)
            control.add_consumer(hostname, destination=[worker])
            queues.append(hostname)
        return queues

    def workers(self):
        """ list of workers

        Returns:
            dict of workers => list of active tasks

        See Also:
            celery Inspect.active()
        """
        return self._inspect.active()

    def queues(self):
        """ list queues

        Returns:
            dict of workers => list of queues

        See Also:
            celery Inspect.active_queues()
        """
        local_q = {gethostname(): [{'name': 'local', 'is_local': True}]}
        celery_qs = self._inspect.active_queues() or {}
        return dict_merge(local_q, celery_qs)

    def labels(self):
        """ list available labels

        Returns:
            dict of workers => list of lables
        """
        return {worker: [q.get('name') for q in queues]
                for worker, queues in self.queues().items()}

    def stats(self):
        """ worker statistics

        Returns:
            dict of workers => dict of stats

        See Also:
            celery Inspect.stats()
        """
        return self._inspect.stats()

    def callback(self, script_name, always=False, **kwargs):
        """ Add a callback to a registered script

        The callback will be triggered upon successful or failed
        execution of the runtime tasks. The script syntax is::

            # script.py
            def run(om, state=None, result=None, **kwargs):
                # state (str): 'SUCCESS'|'ERROR'
                # result (obj): the task's serialized result

        Args:
            script_name (str): the name of the script (in om.scripts)
            always (bool): if True always apply this callback, defaults to False
            **kwargs: and other kwargs to pass on to the script

        Returns:
            self
        """
        success_sig = (self.script(script_name)
                       .task(as_callback=True)
                       .signature(args=['SUCCESS', script_name],
                                  kwargs=kwargs,
                                  immutable=False))
        error_sig = (self.script(script_name)
                     .task(as_callback=True)
                     .signature(args=['ERROR', script_name],
                                kwargs=kwargs,
                                immutable=False))

        if always:
            self._task_default_kwargs['routing']['link'] = success_sig
            self._task_default_kwargs['routing']['link_error'] = error_sig
        else:
            self._require_kwargs['routing']['link'] = success_sig
            self._require_kwargs['routing']['link_error'] = error_sig
        return self


# apply mixins
from omegaml.runtimes.mixins.taskcanvas import canvas_chain, canvas_group, canvas_chord
from omegaml.runtimes.mixins.swagger import SwaggerGenerator

OmegaRuntime.sequence = canvas_chain
OmegaRuntime.parallel = canvas_group
OmegaRuntime.mapreduce = canvas_chord
OmegaRuntime.swagger = SwaggerGenerator.build_swagger



import os

from omegaml import load_class
from omegaml.util import SystemPosixPath


class OmegaTrackingProxy:
    """ OmegaTrackingProxy provides the runtime context for experiment tracking

    Usage:

        Using implied start()/stop() semantics, creating experiment runs::

            with om.runtime.experiment('myexp') as exp:
                ...
                exp.log_metric('accuracy', score)


        Using explicit start()/stop() semantics::

            exp = om.runtime.experiment('myexp')
            exp.start()
            ...
            exp.stop()

    See Also:

        * OmegaSimpleTracker
        * ExperimentBackend
    """
    # test support, does not actually track nesting
    __nested = 0

    def __init__(self, experiment=None, provider=None, runtime=None, implied_run=True):
        self.runtime = runtime
        self.provider = provider
        self._experiment = experiment
        self._implied_run = implied_run
        self._with_experiment = None
        self._tracker = None

    def __getattr__(self, item):
        if hasattr(self.experiment, item):
            return getattr(self.experiment, item)
        return super().__getattribute__(item)

    def create_experiment(self, name, *args, provider=None, **kwargs):
        om = self.runtime.omega
        provider = provider or 'default'
        trackercls = load_class(om.defaults.OMEGA_TRACKING_PROVIDERS.get(provider))
        tracker = trackercls(name, *args, store=om.datasets, model_store=om.models, **kwargs)
        meta = om.models.put(tracker, name, noversion=True)
        return tracker.experiment(name)

    @property
    def experiment(self):
        from omegaml.backends.experiment import ExperimentBackend

        if self._tracker is None:
            om = self.runtime.omega
            fqdn = SystemPosixPath(ExperimentBackend.exp_prefix) / self._experiment
            exp = (om.models.get(str(fqdn), data_store=om.datasets) or
                   self.create_experiment(self._experiment, provider=self.provider))
            self._tracker = exp
        return self._tracker

    def __enter__(self):
        # specify a permanent task argument
        OmegaTrackingProxy.__nested += 1
        self.runtime.require(task=dict(__experiment=self._experiment), always=True)
        self._with_experiment = experiment = self.experiment
        no_active_run = experiment.status() in ('STOPPED', 'PENDING')
        if self._implied_run or no_active_run:
            experiment.start()
            self._implied_run = True if no_active_run else self._implied_run
        return experiment

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._implied_run:
            experiment = self._with_experiment or self.experiment
            experiment.stop()
        self._with_experiment = None
        OmegaTrackingProxy.__nested -= 1
        # remove task argument
        if OmegaTrackingProxy.__nested == 0:
            self.runtime.require(task=dict(__experiment=None), always=True)



from omegaml.backends.basemodel import BaseModelBackend
from omegaml.runtimes.rsystem import rhelper


class RModelBackend(BaseModelBackend):
    KIND = 'model.r'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return str(obj).startswith('r$')

    @property
    def r(self):
        return rhelper()

    def _package_model(self, model, key, tmpfn, **kwargs):
        modelName = model.split('$')[-1]
        self.r.om_save_model(modelName, tmpfn)
        return tmpfn

    def _extract_model(self, infile, key, tmpfn, **kwargs):
        with open(tmpfn, 'wb') as fout:
            fout.write(infile.read())
        obj = self.r.om_load_model(tmpfn, key)
        return RModelProxy(obj)

    def predict(
          self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        rmodel = self.get(modelname)
        result = rmodel.predict(Xname)
        return result


class RModelProxy:
    """ a Python proxy to the R process that runs a model

    This provides the ``model.predict()`` interface for R models so that
    we can use the same semantics for R and python scripts.
    """
    def __init__(self, key):
        self.key = key

    @property
    def r(self):
        return rhelper()

    def predict(self, X_or_name):
        """ call $model.predict()

        Args:
            X_or_name (str): the X object or name of the dataset
        """
        if isinstance(X_or_name, str):
            result = self.r.om_model_predict(self, X_or_name)
        else:
            result = self.r.om_model_predict_py(self, X_or_name)
        return result



from setuptools import setup, find_packages

setup(
    name='envinstall',
    version='',
    packages=find_packages(),
    url='',
    license='',
    author='patrick',
    author_email='',
    description=''
)



import cachetools

from omegaml import defaults as _base_config
from omegaml._version import version
from omegaml.omega import OmegaDeferredInstance
from omegaml.util import load_class, settings, base_loader

# session cache must be defined here globally so can be imported from anywhere
# without causing circular import issues
#: session cache
session_cache = cachetools.cached(cache=cachetools.TTLCache(**_base_config.OMEGA_SESSION_CACHE))

# link implementation
def link(_omega):
    # link a specific implementation lazy loaded at runtime
    global datasets, models, jobs, scripts, runtime, streams, logger, defaults, setup, version
    datasets = _omega.datasets
    models = _omega.models
    jobs = _omega.jobs
    scripts = _omega.scripts
    runtime = _omega.runtime
    streams = _omega.streams
    logger = _omega.logger
    defaults = _omega.defaults
    setup = getattr(_omega, 'setup', setup)
    version = getattr(_omega, 'version', version)


# load base and lazy link
# -- base_loader only loads classes, does not instantiate
# -- deferred instance provides setup to load and link on access
_omega = base_loader(_base_config)
setup = _omega.setup
version = getattr(_omega, 'version', version)
Omega = _omega.Omega
_omega.OmegaDeferredInstance = getattr(_omega, 'OmegaDeferredInstance', OmegaDeferredInstance)
# setup API
#: the :class:`omegaml.store.base.OmegaStore` store for datasets
datasets = _omega.OmegaDeferredInstance(_omega._om, 'datasets')
#: the :class:`omegaml.store.base.OmegaStore` store for models
models = _omega.OmegaDeferredInstance(_omega._om, 'models')
#: the :class:`omegaml.notebook.jobs.OmegaJobs` store for jobs
jobs = _omega.OmegaDeferredInstance(_omega._om, 'jobs')
#: the :class:`omegaml.store.base.OmegaStore` store for scripts
scripts = _omega.OmegaDeferredInstance(_omega._om, 'scripts')
#: the :class:`omegaml.runtimes.runtime.OmegaRuntime` runtime
runtime = _omega.OmegaDeferredInstance(_omega._om, 'runtime')
#: stream helper
streams = _omega.OmegaDeferredInstance(_omega._om, 'streams')
#: the OmegaSimpleLogger for easy log access
logger = _omega.OmegaDeferredInstance(_omega._om, 'logger')
#: the settings object
defaults = _omega.OmegaDeferredInstance(_omega._om, 'defaults')



import os
import pandas as pd
import sys
import unittest
from pandas._testing import assert_frame_equal
from pathlib import Path
from sklearn.linear_model import LinearRegression
from types import ModuleType
from unittest.mock import patch

from omegaml import Omega
from omegaml.mixins.store.imexport import ObjectImportExportMixin, OmegaExportArchive, OmegaExporter
from omegaml.tests.util import OmegaTestMixin


class ImportExportMixinTests(OmegaTestMixin, unittest.TestCase):
    def setUp(self):
        self.om = om = Omega()
        self.om_restore = self.om['restore']
        for omx in (self.om, self.om_restore):
            self._apply_store_mixin(omx)
        self.clean()
        self.clean(bucket='restore')
        with OmegaExportArchive('/tmp/test', None) as arc:
            arc.clear()

    def _apply_store_mixin(self, omx):
        for store in (omx.datasets, omx.models, omx.jobs.store, omx.scripts, omx.streams):
            store.register_mixin(ObjectImportExportMixin)

    def test_dataframe_export(self):
        om = self.om
        # export
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'mydf', append=False)
        om.datasets.to_archive('mydf', '/tmp/test')
        # import
        om.datasets.drop('mydf', force=True)
        om.datasets.from_archive('/tmp/test', 'mydf')
        xdf = om.datasets.get('mydf')
        assert_frame_equal(xdf, df)

    def test_model_export_singleversion(self):
        om = self.om
        om_restore = self.om_restore
        model = LinearRegression()
        model.coef_ = 1
        om.models.put(model, 'mymodel')
        om.models.to_archive('mymodel', '/tmp/test')
        om.models.drop('mymodel', force=True)
        om.models.from_archive('/tmp/test', 'mymodel')
        mdl = om.models.get('mymodel')
        self.assertIsInstance(mdl, LinearRegression)
        om_restore.models.from_archive('/tmp/test', 'mymodel')
        mdl = om_restore.models.get('mymodel')
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 1)

    def test_model_export_multiversion(self):
        om = self.om
        om_restore = self.om_restore
        model = LinearRegression()
        model.coef_ = 1
        model = LinearRegression()
        model.coef_ = 2
        om.models.put(model, 'mymodel', tag='latest')
        om.models.put(model, 'mymodel', tag='version2')
        om.models.to_archive('mymodel', '/tmp/test')
        om.models.drop('mymodel', force=True)
        om.models.from_archive('/tmp/test', 'mymodel')
        mdl = om.models.get('mymodel')
        self.assertIsInstance(mdl, LinearRegression)
        om_restore.models.from_archive('/tmp/test', 'mymodel')
        mdl = om_restore.models.get('mymodel')
        # we expect the latest version
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 2)
        self.assertNotIn('versions', om_restore.models.metadata('mymodel').attributes)

    def test_model_multiple_versioned_export(self):
        om = self.om
        om_restore = self.om_restore
        model = LinearRegression()
        model.coef_ = 1
        om.models.put(model, 'mymodel', tag='version1')
        model.coef_ = 2
        om.models.put(model, 'mymodel', tag='version2')
        print(om.models.get('mymodel@version1').coef_)
        print(om.models.get('mymodel@version2').coef_)
        # save specific versions
        om.models.to_archive('mymodel@version1', '/tmp/test')
        om.models.to_archive('mymodel@version2', '/tmp/test')
        om.models.drop('mymodel', force=True)
        # restore specific versions
        om_restore.models.from_archive('/tmp/test', 'mymodel@version1')
        om_restore.models.from_archive('/tmp/test', 'mymodel@version2')
        # check we can access specific versions, as restored
        mdl = om_restore.models.get('mymodel@version1')
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 1)
        mdl = om_restore.models.get('mymodel@version2')
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 2)
        # we don't expect a versioned base model
        self.assertNotIn('mymodel', om.models.list())
        self.assertNotIn('mymodel@version1', om.models.list())
        self.assertNotIn('mymodel@version2', om.models.list())
        # promote from restore bucket into a versioned model
        om_restore.models.promote('mymodel@version1', om.models)
        om_restore.models.promote('mymodel@version2', om.models)
        # check the model is actually versioned
        mdl = om.models.get('mymodel@latest')
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 2)
        self.assertEqual(om.models.list(hidden=True), ['mymodel'])
        self.assertNotIn('mymodel@version1', om.models.list())
        self.assertNotIn('mymodel@version2', om.models.list())
        self.assertEqual(len(om.models.revisions('mymodel')), 2)
        # check we can access the versioned model
        mdl = om.models.get('mymodel@version1')
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 1)
        mdl = om.models.get('mymodel@version2')
        self.assertIsInstance(mdl, LinearRegression)
        self.assertEqual(mdl.coef_, 2)

    def test_jobs_export(self):
        om = self.om
        code = "print('hello world')"
        om.jobs.create(code, 'mynb')
        om.jobs.to_archive('mynb', '/tmp/test')
        om.jobs.drop('mynb', force=True)
        om.jobs.from_archive('/tmp/test', 'mynb')
        job = om.jobs.get('mynb')
        print(job)

    def test_script_export(self):
        om = self.om
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'helloworld')
        om.scripts.to_archive('helloworld', '/tmp/test')
        om.scripts.drop('helloworld', force=True)
        om.scripts.from_archive('/tmp/test', 'helloworld')
        mod = om.scripts.get('helloworld')
        self.assertIsInstance(mod, ModuleType)
        self.assertTrue(hasattr(mod, 'run'))

    def test_streams_export(self):
        om = self.om
        stream = om.streams.get('mystream')
        meta = om.streams.metadata('mystream')
        meta.attributes['some_data'] = 'foo'
        meta.save()
        om.streams.to_archive('mystream', '/tmp/test')
        om.streams.drop('mystream', force=True)
        om.streams.from_archive('/tmp/test', 'mystream')
        meta = om.streams.metadata('mystream')
        self.assertEqual(meta.attributes.get('some_data'), 'foo')

    def test_compress(self):
        om = self.om
        om_restore = self.om_restore
        # export
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'mydf', append=False)
        arc = om.datasets.to_archive('mydf', '/tmp/test')
        # create an archive
        arcfile = arc.compress()
        # -- the archive path is deleted
        self.assertFalse(Path(arc.path).exists())
        # -- there is an timestamped archive file (tgz)
        self.assertTrue(Path(arcfile).is_file())
        self.assertTrue(Path(arcfile).name.endswith('.tgz'))
        # import from the tgz archive
        om_restore.datasets.drop('mydf', force=True)
        om_restore.datasets.from_archive(arcfile, 'mydf')
        xdf = om_restore.datasets.get('mydf')
        assert_frame_equal(xdf, df)

    def test_runtime_exporter_export_import(self):
        om = self.om
        om_restore = self.om_restore
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'mydf', append=False)
        OmegaExporter(om).to_archive('/tmp/test', ['data/mydf'])
        imported = OmegaExporter(om_restore).from_archive('/tmp/test')
        meta = om_restore.datasets.metadata('mydf')
        self.assertIn('mydf', om_restore.datasets.list())
        self.assertEqual(imported, [meta])

    def test_runtime_exporter_export_import_compressed(self):
        om = self.om
        om_restore = self.om_restore
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'mydf', append=False)
        arcfile = OmegaExporter(om).to_archive('/tmp/test', ['data/mydf'], compress=True)
        with self.assertRaises(FileNotFoundError):
            OmegaExporter(om_restore).from_archive('/tmp/test')
        imported = OmegaExporter(om_restore).from_archive(arcfile)
        meta = om_restore.datasets.metadata('mydf')
        self.assertIn('mydf', om_restore.datasets.list())
        self.assertEqual(imported, [meta])

    def test_runtime_exporter_export_promote(self):
        om = self.om
        om_restore = self.om_restore
        # dataset
        df = pd.DataFrame({'x': range(10)})
        om.datasets.put(df, 'mydf', append=False)
        # models
        model = LinearRegression()
        model.coef_ = 1
        om.models.put(model, 'mymodel', tag='version1')
        model.coef_ = 2
        om.models.put(model, 'mymodel', tag='version2')
        # job
        code = "print('hello')"
        om.jobs.create(code, 'myjob')
        # export
        OmegaExporter(om).to_archive('/tmp/test', ['data/mydf',
                                                   'models/mymodel@version1',
                                                   'models/mymodel@version2',
                                                   'jobs/myjob'])
        # import
        # -- mock temp bucket retrieval in order to test promotion
        temp_bucket = om[OmegaExporter._temp_bucket]
        self._apply_store_mixin(temp_bucket)
        with patch.object(om_restore, '_get_bucket') as meth:
            meth.return_value = temp_bucket
            OmegaExporter(om_restore).from_archive('/tmp/test',
                                                   pattern='data/.*|models/.*',
                                                   promote_to=om_restore)
            self.assertIn('mydf', om_restore.datasets.list())
            self.assertIn('mymodel', om_restore.models.list())
            self.assertEqual(len(om_restore.models.revisions('mymodel')), 2)
            # check model versions are as expected
            # -- latest
            mdl = om_restore.models.get('mymodel')
            self.assertIsInstance(mdl, LinearRegression)
            self.assertEqual(mdl.coef_, 2)
            # previous
            mdlv1 = om_restore.models.get('mymodel@version1')
            self.assertIsInstance(mdlv1, LinearRegression)
            self.assertEqual(mdlv1.coef_, 1)
            # check jobs were not restored yet
            self.assertEqual(om_restore.jobs.list(), [])
            # restore jobs explicitly
            # -- note jobs promotion is not supported (pending #218)
            OmegaExporter(om_restore).from_archive('/tmp/test',
                                                   pattern='jobs/.*')
            self.assertEqual(om_restore.jobs.list(), ['myjob.ipynb'])
            # ensure models were not touched
            self.assertEqual(len(om_restore.models.revisions('mymodel')), 2)


if __name__ == '__main__':
    unittest.main()



"""
omega runtime script tasks
"""
from __future__ import absolute_import

import datetime
from celery import shared_task

from omegaml.celery_util import OmegamlTask
from omegaml.util import json_dumps_np


class ScriptTask(OmegamlTask):
    abstract = True

    def _get_script_metadata(self, args, kwargs):
        om = self.om
        scriptname, *args = args
        meta = om.scripts.metadata(scriptname)
        return meta

    def on_success(self, retval, task_id, args, kwargs):
        meta = self._get_script_metadata(args, kwargs)
        attrs = meta.attributes
        attrs['state'] = 'SUCCESS'
        attrs['task_id'] = task_id

        if not kwargs:
            pass
        else:
            attrs['last_run_time'] = kwargs.get('run_at')
            attrs['next_run_time'] = kwargs.get('next_run_time')

        meta.attributes = attrs
        meta.save()

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        meta = self._get_script_metadata(args, kwargs)
        attrs = meta.attributes
        attrs['state'] = 'FAILURE'
        attrs['task_id'] = task_id

        if not kwargs:
            pass
        else:
            attrs['last_run_time'] = kwargs.get('run_at')
            attrs['next_run_time'] = kwargs.get('next_run_time')

        meta.attributes = attrs
        meta.save()


class CallbackScriptTask(ScriptTask):
    def _get_script_metadata(self, args, kwargs):
        om = self.om
        *args, scriptname = args
        meta = om.scripts.metadata(scriptname)
        return meta


@shared_task(bind=True, base=ScriptTask)
def run_omega_script(self, scriptname, *args, **kwargs):
    """
    runs omegaml job
    """
    # setup options
    SERIALIZER = {
        'json': json_dumps_np,
        'python': lambda v: v,
    }
    format = self.system_kwargs.get('__format') or 'json'
    # call and measure time
    dtstart = datetime.datetime.now()
    try:
        result = (self.get_delegate(scriptname, kind='scripts', pass_as='data_store')
                  .perform('run', scriptname, *self.delegate_args[1:], om=self.om, **self.delegate_kwargs))
    except Exception as e:
        import traceback
        if kwargs.get('__traceback') or self.logging[-1].lower() == 'debug':
            result = "".join(traceback.TracebackException.from_exception(e).format())
        else:
            result = repr(e)
        exception = True
    else:
        exception = False
    dtend = datetime.datetime.now()
    duration = dtend - dtstart
    # prepare result
    data = {
        'script': scriptname,
        'args': self.delegate_args[1:],
        'kwargs': self.delegate_kwargs,
        'result': result,
        'runtimes': float(duration.seconds) + duration.microseconds / float(1e6),
        'started': dtstart.isoformat(),
        'ended': dtend.isoformat(),
    }
    data = SERIALIZER[format](data)
    if exception:
        raise RuntimeError(data)
    return data


@shared_task(bind=True, base=CallbackScriptTask)
def run_omega_callback_script(self, *args, **kwargs):
    """
    runs omegaml job
    """
    # callbacks have the result first, then the original arguments
    # -- thus scriptname is the last(!) argument
    # -- https://docs.celeryq.dev/en/stable/userguide/canvas.html#callbacks
    if len(args) >= 3:
        results = args[0:-2]
        state = args[-2]
        scriptname = args[-1]
    else:
        results = args[0]
        state = 'SUCCESS'
        scriptname = args[-1]
    # include hidden (__key=value) kwargs, e.g. for logging, bucket, experiment
    # so that the run_omega_script task context sets up a proper omega instance
    full_kwargs = kwargs
    full_kwargs.update(self.system_kwargs)
    return run_omega_script(scriptname, state=state, results=results, **full_kwargs)



from hashlib import md5
from mongoengine import DoesNotExist

from omegaml.store import OmegaStore


class StreamsProxy(OmegaStore):
    """
    A straight forward adapter to minibatch streams

    Note backends or mixins are disabled for StreamsProxy.

    Usage:
        # insert data to a stream
        om.streams.put(data, 'name')

        # to increase performance, use the stream.append() method
        stream = om.streams.get('name')
        stream.append(data)

        # get a minibatch stream. note it always exists
        om.streams.get('name')

        # get the buffer of a stream
        om.streams.get('name').buffer()

        # get a stream's metadata
        om.streams.metadata('name')

        # get a streaming emitter function
        emitter = om.streams.get('name', lazy=True, **kwargs)
        emitter = om.streams.getl('name', **kwargs)
        # run with a callable
        # -- the callable will be called in parallel
        emitter(lambda window: ...)

        # specify the size (N) or interval (seconds) of the window
        emitter = om.streams.getl('name', size=10, **kwargs)
        emitter = om.streams.getl('name', interval=5, **kwargs)

    See Also:
        minibatch package for details
    """

    # TODO move this implementation to a proper OmegaStore backend or mixin
    # this is a OmegaStore so we can have Metadata support
    # it is called the StreamsProxy because it behaves more like a Mixin
    # rationale for not using a proper Backend on om.datasets is that
    # we wanted the streams.put() method to be really fast, without any lookups
    # It may be worthwhile to reconsider an just make streams.get().append()
    # the high performance alternative
    def __init__(self, prefix=None, bucket=None, defaults=None, mongo_url=None, **kwargs):
        super().__init__(mongo_url=mongo_url, prefix=prefix, bucket=bucket, defaults=defaults)
        cnx_kwargs = self.defaults.OMEGA_MONGO_SSL_KWARGS
        cnx_kwargs.update(kwargs)
        self._stream_kwargs = dict(url=self.mongo_url, **cnx_kwargs)

    def _mixins_conditional(self, cls, obj):
        # only allow mixins that do not interfere with streams get/put
        return cls.supports(obj, prefix='streams/') if hasattr(cls, 'supports') else False

    def register_backends(self):
        # TODO enable custom backends
        # disabled to avoid interference with custom get(), put()
        pass

    def _qualified_stream(self, name, *args, **kwargs):
        return f'{self.bucket}.{self.prefix}.{name}.stream'

    def get(self, name, lazy=False, **kwargs):
        import minibatch as mb
        meta = self.metadata(name)
        if meta is None:
            kind_meta = {
                'stream': self._qualified_stream(name),
                'kwargs': {
                    'batchsize': kwargs.get('batchsize')
                },
            }
            meta = self.make_metadata(name,
                                      'stream.minibatch',
                                      prefix=self.prefix,
                                      bucket=self.bucket,
                                      kind_meta=kind_meta).save()
        stream_name = meta.kind_meta['stream']
        stream_kwargs = dict(self._stream_kwargs)
        stream_kwargs.update(meta.kind_meta.get('kwargs', {}))
        if not lazy:
            stream = mb.stream(stream_name, **kwargs, **self._stream_kwargs)
        else:
            stream = mb.streaming(stream_name, **kwargs, cnx_kwargs=self._stream_kwargs)
        return stream

    def _cached_get(self, name, reload=False, _cachex=None):
        if reload:
            # force to avoid cache
            return self.get(name, _cachex=md5().hexdigest())
        return self.get(name)

    def getl(self, name, **kwargs):
        return self.get(name, lazy=True, **kwargs)

    def drop(self, name, force=False):
        meta = self.metadata(name)
        if meta is None and not force:
            raise DoesNotExist()
        stream = self.get(name)
        stream.buffer().delete()
        stream.delete()
        meta.delete() if meta is not None else None
        return True

    def _recreate(self, name):
        self.drop(name)
        return self.get(name)

    def put(self, data, name, append=True):
        (self._cached_get(name) if append else self._recreate()).append(data)



# TODO implement drop for revisions!
from datetime import datetime

import pandas as pd
from tqdm import tqdm

from omegaml.util import tryOr


class DataRevisionMixin:
    """ Enable automatically versioned dataframes in om.datasets

    Works by storing a current version of the dataframe with all changes applied
    as the most recent revision. All changes are stored separately in sequence,
    such that any change or previous "as at" version can be retrieved.

    Usage:
        # initial
        om.datasets.put(df, 'mydf', revisions=True)

        # changesets -- will merge mydf by index, and keep track of updates
        om.datasets.put(updates_df, 'mydf')
        om.datasets.put(updates_df, 'mydf', revision_dt=<dt>)
        om.datasets.put(updates_df, 'mydf', tag='tagname')

        # changesets can be used to delete
        # -- delete all by index
        om.datasets.put(updates_df, 'mydf', delete=True)
        # -- delete by specific rows
        updates_df['_delete_'] = True
        om.datasets.put(updates_df, 'mydf')

        # retrieve latest revision, all changes applied
        # -- existing rows are updated by matching the dataframe's index
        # -- new rows are inserted
        # -- deleted rows are removed
        om.datasets.get('mydf')

        # retrieve as at revision
        om.datasets.get('mydf', revision=N|datetime|tag)

        # retrieve changes only
        # -- result contains additional columns, see trace revisions below
        om.datasets.get('mydf', changeset=N)

        # trace revisions - similar to changeset, but apply all revisions
        # -- result contains additional columns
        #    - _om#revision - the sequence this row was last applied from
        #    - _delete_ - the delete flag, if True row was deleted in change set (note the
        #                 deletion is _not_ applied upon trace_revisions=True)
        om.datasets.get('mydf', trace_revisions=True)

    Metadata:
        meta.kind_meta['revisions'] = {
            'seq': N, # the most recent revision sequence, starting at 0
            'name': '.revisions.dataset', #the name of the revision dataset
            'changes': [{
               'dt': <timestamp>, # the creation timestamp of the change
               'seq': N, # the sequence of the revision
               'tags': [], # list of tags to identify this revision
            }, ...]
        }

    Notes:
        * this is currently implemented for pd.DataFrames only, however the mixin
          uses a revision protocol that is independent of the underlying storage
        * to implement revision support for another storage, override the _make_upsert_fn,
          _get_revisions, _apply_changeset methods. In a minimal implementation, _make_upsert_fn
          is the only method required as all other methods leverage the OmegaStore semantics.

    """

    @classmethod
    def supports(cls, store, **kwargs):
        return store.prefix == 'data/'

    def _has_revisions(self, name, revisions=False):
        # return True if revisions exist for dataset name, or if requested
        meta = self.metadata(name)
        revisions = revisions or (meta is not None and 'revisions' in meta.kind_meta)
        return revisions

    def _build_revision(self, df, name, append=True, revision_dt=None, tag=None, delete=False, **kwargs):
        """ build a new revision

        Creates and updates a 'revisions' entry in the datasets meta.kind_meta, keeping
        track of all changes:

        meta.kind_meta['revisions'] = {
           'seq': N, # the most recent revision, 0-indexed
           'name': revname # the name of the revision collection
           'changes': [{
              'dt': <datetime>,
              'seq': N,
              'tags': [str, ...],
              'delete': True|False,
           }, ...]
        }
        """
        # build a new revision
        # -- set revision dataset
        revname = '.revisions.{name}'.format(**locals())
        # -- deal with replacement, dropping dataset and all revisions
        if not append:
            super().drop(name, force=True)
            super().drop(revname, force=True)
        # build revision metadata
        if '_delete_' not in df.columns:
            df['_delete_'] = delete
        meta = self.metadata(name)
        if meta is not None:
            # _calculate and record revision
            revision = meta.kind_meta['revisions'].get('seq', 0) + 1
            df['_om#revision'] = revision
        else:
            revision = 0
            df['_om#revision'] = revision
            meta = super().put(df, name, append=False, **kwargs)
        revisions = meta.kind_meta.setdefault('revisions', {})
        changesets = revisions.setdefault('changes', [])
        revisions.update({
            'seq': revision,
            'name': revname
        })
        changesets.append({
            'dt': revision_dt or datetime.utcnow(),
            'seq': revision,
            'tags': [tag] if tag else [],
            'delete': delete,
        })
        meta.save()
        # -- record all changes to revision dataset
        self._store_changeset(df, revname, append=append)
        return meta

    def _store_changeset(self, df, revname, append=True, **kwargs):
        return super().put(df, revname, append=append)

    def _retrieve_revision(self, name, revision=-1, changeset=None, trace_revisions=False, **kwargs):
        # retrieve a revision or a specific changeset
        # -- determine revision to get
        meta = self.metadata(name)
        revname = meta.kind_meta['revisions']['name']
        latest_revision = meta.kind_meta['revisions']['seq']
        changes = meta.kind_meta['revisions']['changes']
        # -- return changeset if requested
        if changeset is not None:
            data = super().get(revname, filter={'_om#revision': changeset}, **kwargs)
            return self._clean(data, trace_revisions=trace_revisions)
        if isinstance(revision, int):
            # get revision by sequence
            revision = revision if revision is not None else latest_revision
            revision = min(revision, latest_revision)
            if revision < -1:
                revision = changes[revision]['seq']
        elif isinstance(revision, datetime):
            # get revision by nearest date ("as before or latest at")
            requested = revision
            revision = changes[0]['seq']  # earliest revision by default
            for cs in changes:
                if cs['dt'] > requested:
                    break
                revision = cs['seq']
            else:
                # we did find any before or at, return latest
                revision = latest_revision
        else:
            # get revision by tag
            changes = meta.kind_meta['revisions']['changes']
            requested = revision
            for cs in changes:
                if requested in cs['tags']:
                    revision = cs['seq']
                    break
            else:
                return None
        # -- latest revision is simply the current dataset
        if not trace_revisions and (revision == -1 or revision == latest_revision):
            data = super().get(name, **kwargs)
        else:
            # -- get all changesets up to requested revision
            data = self._apply_changesets(revname, revision, trace_revisions=trace_revisions, **kwargs)
        self._clean(data, trace_revisions=trace_revisions)
        return data

    def _apply_changesets(self, revname, revision, trace_revisions=False, **kwargs):
        # get base revision and apply all changesets up to requested by merging in sequence
        # -- merge
        base = super().get(revname, filter={'_om#revision': 0}, **kwargs)
        base_types = base.dtypes
        idx_type = base.index.dtype
        for rev in range(1, revision + 1):
            changes = super().get(revname, filter={'_om#revision': rev}, **kwargs)
            if changes is None:
                continue
            # apply upserts, step 1
            base = base.merge(changes,
                              how='outer',
                              left_index=True,
                              right_index=True,
                              suffixes=(None, '_r_'))
            # apply upserts, step 2
            revcols = []
            for col in [c for c in base.columns if not c.endswith('_r_')]:
                revcol = col + '_r_'
                base[col] = base[revcol].fillna(base[col]).astype(base_types[col])
                revcols.append(revcol)
            base.index = base.index.astype(idx_type)
            if revcols:
                base.drop(columns=revcols, inplace=True)
            # apply deletions
            if not trace_revisions:
                flt_delete = base['_delete_'] == True  # noqa
                base = base[~flt_delete]
        return base

    def _make_upsert_fn(self, name, delete=False):
        collection = self.collection(name)

        def upsert(obj, store, name, chunksize=1):  # noqa
            from pymongo import DeleteOne, UpdateOne

            ops_updates = []
            ops_deletions = []

            for i, row in tqdm(obj.iterrows(), total=len(obj)):
                data = row.to_dict()
                # om rowid is re-added on insert only to preserve existing row id
                del data['_om#rowid']
                updates = {
                    '$set': data,
                    '$setOnInsert': {
                        '_om#rowid': row['_om#rowid']
                    }
                }
                key = {k: v for k, v in data.items() if k.startswith('_idx#')}
                if data.get('_delete_', delete):
                    # collection.delete_one(key)
                    ops_deletions.append(DeleteOne(key))
                else:
                    # collection.update_one(key, updates, upsert=True)
                    ops_updates.append(UpdateOne(key, updates, upsert=True))

            # do all updates in one step
            collection.bulk_write(ops_deletions + ops_updates)

        return upsert

    def _clean(self, df, trace_revisions=False):
        if not trace_revisions:
            df.drop(columns=['_om#revision', '_delete_'], inplace=True)
        return df

    def put(self, df, name, revisions=False, tag=None, revision_dt=None, trace_revisions=False, **kwargs):
        """ store a dataset revision

        Args:
            df (pd.DataFrame): a dataframe
            name (str): name of dataset
            append (bool): whether to append or replace the dataset, by default append, defaults to True
            delete (bool): whether to upsert changes or to delete by index, defaults to False i.e. upsert
            revisions (bool): if True, keep revisions. defaults to False
            revision_dt (datetime): if specified set this as the revision datetime
            tag (str): optional, if specified record this revision given the tag
            trace_revisions (bool): optional, if True return information on applied revisions (deletions
              are flagged as the _delete_ flag, not applied)
        """
        if not self._has_revisions(name, revisions=revisions):
            return super().put(df, name, **kwargs)
        # revisions apply
        append = kwargs.pop('append', True)
        delete = kwargs.pop('delete', False)
        meta = self._build_revision(df, name, append=append, tag=tag, delete=delete,
                                    revision_dt=revision_dt, **kwargs)
        if append:
            # _fast_insert is a callback to process the upserts
            super().put(df, name, _fast_insert=self._make_upsert_fn(name, delete=delete))
        self._clean(df, trace_revisions=trace_revisions)
        return meta

    def get(self, name, revision=-1, changeset=None, trace_revisions=False, **kwargs):
        """ retrieve a dataset with revisions

        Retrieve a specific revision, including all changes applied until that point. If
        revision is -1 (latest), the current dataset is returned. For any other revision,
        all changes up to this point are applied by merging the first revision with all
        later changes. If a changeset is requested only this changeset is returned. Revisions
        can be specified as a zero-indexed sequence number.

        Args:
            name (str): dataset name
            revision (int|datetime|str): optional, 0-indexed revision sequence (0=first) or -1 negative index
              since last change (-1 = latest), or previous and up to changes before datetime, or
              a tag name. Defaults to -1, returning the latest version of the data
            changeset: optional, if given, specifies the 0-indexed changeset sequence, if specified
              changeset is returned
            trace_revisions (bool): optional, if True return information on applied revisions (deletions
              are flagged as the _delete_ flag, not applied)
        """
        # we pass revisions=None to enforce decision by meta data
        if not self._has_revisions(name, revisions=None):
            return super().get(name, **kwargs)
        # check if revision is a datetime value
        if isinstance(revision, str):
            revision = tryOr(lambda: pd.to_datetime(revision).to_pydatetime(), revision)
        data = self._retrieve_revision(name,
                                       revision=revision,
                                       changeset=changeset,
                                       trace_revisions=trace_revisions,
                                       **kwargs)
        return data

    def revisions(self, name):
        meta = self.metadata(name)
        changes = meta.kind_meta['revisions']['changes']
        return pd.DataFrame(changes)



from __future__ import annotations

import json
import re
import tarfile
from bson.json_util import dumps as bson_dumps, loads as bson_loads
from datetime import datetime
from pathlib import Path
from shutil import rmtree

from omegaml.client.util import AttrDict
from omegaml.documents import Metadata
from omegaml.mixins.store.promotion import PromotionMixin
from omegaml.omega import Omega
from omegaml.store import OmegaStore
from omegaml.util import IterableJsonDump, load_class, SystemPosixPath


class ObjectImportExportMixin:
    """ Provide generic import() and export() methods """

    @classmethod
    def supports(cls, store, **kwargs):
        return store.prefix in ('data/', 'jobs/', 'models/', 'scripts/', 'streams/')

    def to_archive(self, name, local, fmt='omega', **kwargs):
        backend = self.get_backend(name)
        if hasattr(backend, 'export'):
            return backend.export(name)
        meta = self.metadata(name, raw=True)
        assert meta is not None, f"{name} is not in {self.prefix}"
        if not isinstance(local, OmegaExportArchive):
            archive = OmegaExporter.archive(local, store=self, fmt=fmt)
        else:
            archive = local
        with archive as arc:
            arc.add(meta, asname=name, store=self)
        return arc

    def from_archive(self, local, name, fmt='omega', **kwargs):
        self.drop(name, force=True)
        meta = self._load_metadata(name)
        if not isinstance(local, OmegaExportArchive):
            archive = OmegaExporter.archive(local, store=self, fmt=fmt)
        else:
            archive = local
        with archive as arc:
            meta = arc.extract(name, meta, asname=meta.name, store=self)
        return meta

    def _load_metadata(self, name, attributes=None, gridfile=None):
        meta = self.metadata(name, raw=True)
        return meta or self._make_metadata(
            name=name,
            prefix=self.prefix,
            bucket=self.bucket,
            attributes=attributes,
            gridfile=gridfile)


class OmegaExportArchive:
    def __init__(self, path, store=None):
        self.path = Path(path)
        self.store = store
        self.manifest = self._read_manifest()
        self._with_arc = None

    def __enter__(self, compress=False):
        self._with_arc = arc = self.decompress()
        arc.manifest = arc._read_manifest()
        return arc

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.is_readonly:
            self.close()

    @property
    def is_readonly(self):
        return self is not self._with_arc

    def _writable(fn):
        def inner(self, *args, **kwargs):
            if not self.is_readonly:
                return fn(self, *args, **kwargs)
            else:
                raise ValueError('archive is not writable')

        return inner

    @_writable
    def clear(self):
        if self.path.exists():
            rmtree(self.path)
        self.manifest = self._read_manifest()

    def compress(self):
        # create a timestamped path that does not contain :
        # -- due to https://superuser.com/a/1720174
        dt = str(datetime.utcnow().isoformat()).replace(':', '')
        tfn = self.path.parent / f'{self.path.name}-{dt}.tgz'
        Path(tfn).unlink(missing_ok=True)
        with tarfile.open(tfn, 'w:gz') as tar:
            tar.add(self.path, arcname=self.path.name, recursive=True)
        self.clear()
        return tfn

    def decompress(self):
        if self.path.is_file():
            target = self.path.parent / self.path.name.replace('.tgz', '')
            with tarfile.open(self.path, 'r:gz') as tar:
                tar.extractall(target)
            # the first entry in the archive is the actual archive contents
            basename = list(target.iterdir())[0]
            arc = self.__class__(basename, self.store)
        else:
            arc = self
        return arc

    @property
    def members(self):
        return self.manifest['members'].keys()

    @_writable
    def add(self, meta, asname=None, store=None):
        store = store or self.store
        name = asname or meta.name
        meta_dict = meta.to_dict()
        # prepare local paths
        lpaths = self._local_paths(self.path, name, store)
        # data in gridfile
        data = meta.gridfile.read()
        if data is not None:
            lpaths.gridfile.write_bytes(data)
            meta_dict['gridfile'] = lpaths.gridfile.name  # basename
        # data in collection
        if meta.collection:
            def remove_id(obj):
                obj.pop('_id', None)
                return obj

            data = store.collection(name).find()
            with lpaths.collection.open('w') as fout:
                IterableJsonDump.dump(data, fout,
                                      transform=remove_id,
                                      default=bson_dumps)
        # metadata
        lpaths.meta.write_text(bson_dumps(meta_dict))
        self.manifest['members'][self._manifest_key(name, store)] = lpaths.key

    @_writable
    def close(self):
        self._write_manifest()

    def extract(self, name, meta, asname=None, store=None):
        store = store or self.store
        lpaths = self._local_paths(self.path, name, store, expect_exist=True)
        with lpaths.meta.open('r') as fin:
            meta_dict = bson_loads(fin.read())
            meta.attributes.update(meta_dict['attributes'])
            meta.kind_meta.update(meta_dict['kind_meta'])
            meta.kind = meta_dict['kind']
        if lpaths.gridfile.exists():
            with lpaths.gridfile.open('rb') as fin:
                file_backend = store.get_backend_byobj(fin)
                meta.gridfile = file_backend._store_to_file(store, fin, lpaths.key)
        if lpaths.collection.exists():
            with lpaths.collection.open('r') as fin:
                data = bson_loads(fin.read())
                collection = store.collection(name)
                collection.insert_many(data)
                meta.collection = collection.name
        if asname:
            meta.name = asname
        return meta.save()

    def _manifest_key(self, name, store):
        return f'{store.prefix}{name}'

    def _local_paths(self, local, name, store, expect_exist=False):
        manifest_key = self._manifest_key(name, store)
        in_manifest = manifest_key in self.manifest['members']
        if not in_manifest and not expect_exist:
            local_key = SystemPosixPath(store.prefix) / store.object_store_key(name, 'omx')
        elif in_manifest:
            local_key = self.manifest['members'][manifest_key]
        else:
            raise ValueError(f'{name} expected to exist, but is not a member in archive {self.path}')
        local_dir = Path(local) / local_key
        local_meta = Path(local_dir) / 'metadata.json'
        local_gridfile = Path(local_dir) / 'gridfile.bin'
        local_collection = Path(local_dir) / 'collection.json'
        local_dir.mkdir(parents=True, exist_ok=True)
        return AttrDict(dir=local_dir,
                        key=str(local_key),
                        meta=local_meta,
                        gridfile=local_gridfile,
                        collection=local_collection)

    def _read_manifest(self):
        manifest_path = self.path / 'manifest.json'
        if manifest_path.exists():
            with manifest_path.open('r') as fin:
                manifest = json.loads(fin.read())
        else:
            manifest = {
                'members': {},
                'format': 'omega',
            }
        return manifest

    def _write_manifest(self):
        self.path.mkdir(exist_ok=True)
        with (self.path / 'manifest.json').open('w') as fout:
            json.dump(self.manifest, fout)


class OmegaExporter:
    ARCHIVERS = {
        'omega': 'omegaml.mixins.store.imexport.OmegaExportArchive',
    }
    _temp_bucket = '__exporter' # used as the import target and source for promotion

    def __init__(self, omega):
        self.omega = omega

    def to_archive(self, path, objects=None, fmt='omega', compress=False, progressfn=None):
        obj: Metadata | str
        store: ObjectImportExportMixin | OmegaStore
        if not objects:
            objects = self.omega.list(raw=True)
        with OmegaExporter.archive(path, fmt=fmt) as arc:
            arc.clear()
            for obj in objects:
                progressfn(obj) if progressfn else None
                if isinstance(obj, Metadata):
                    store = self.omega.store_by_meta(obj)
                    store.to_archive(obj.name, arc)
                elif isinstance(obj, str):
                    try:
                        prefix, pattern = obj.split('/', 1)
                    except ValueError:
                        prefixes = [s.prefix for s in self.omega._stores]
                        raise ValueError(f'Cannot parse {obj}. Specify objects as prefix/name, prefix is one of {prefixes}')
                    store = self.omega.store_by_prefix(f'{prefix}/')
                    # if the pattern does not match in list, use it as a name
                    # e.g. mymodel@version1 will not show in list()
                    objects = store.list(pattern=pattern) or [pattern]
                    for objname in objects:
                        store.to_archive(objname, arc)
        if compress:
            archive_path = arc.compress()
        else:
            archive_path = path
        return archive_path

    def from_archive(self, path, pattern=None, fmt='omega', promote=False,
                     promote_to: Omega=None, progressfn=None):
        # for promotion, use a temp bucket for import and promotion source
        promote = promote or (promote_to is not None)
        promote_to = None if not promote else (promote_to or self.omega)
        omega = self.omega if not promote else promote_to[self._temp_bucket]
        store: ObjectImportExportMixin
        imported = []
        pattern = pattern.replace('datasets/', 'data/') if pattern else pattern
        if not Path(path).exists():
            raise FileNotFoundError(path)
        with OmegaExporter.archive(path, fmt=fmt) as arc:
            for member in arc.members:
                progressfn(member) if progressfn else None
                if pattern and not re.match(pattern, member):
                    continue
                prefix, name = member.split('/', 1)
                store = omega.store_by_prefix(f'{prefix}/')
                meta = store.from_archive(arc, name)
                if promote:
                    store: PromotionMixin
                    to_store = promote_to.store_by_prefix(f'{prefix}/')
                    meta = store.promote(name, to_store)
                imported.append(meta)
        return imported

    @classmethod
    def archive(cls, path, store=None, fmt='omega') -> OmegaExportArchive:
        archiver = load_class(cls.ARCHIVERS[fmt])
        archive = archiver(path, store)
        return archive



""" R system runtime support

This implements omegaml's R interface

How it works:

    Using the R:reticulate package omegaml is transparently available
    from within R and RStudio. For the runtime we start the omegaml
    worker (celery) using reticulate (see omworker.R).

    Storage:

        The RModelBackend implements the models storage backend to save and
        load models by providing a facade to the respective R functions.
        Model persistence works the same as with Python models, namely the
        backend generates a temporary path for model saving, then
        calls the respective R function for model saving (see omegaml.R). The
        saved file is then transferred to the storage's filesystem. Model loading
        works the other way around. Note that the actual R model object is not
        converted to Python, instead a uuid-like random name is used to store the
        object in R's OmegaEnv. R code can get the actual object by using the
        rmodel() helper function.

    Runtime:

        The runtime is the omegaml celery worker, started from within an R session.
        To start the worker use the cli:

        $ om runtime celery rworker

        This starts the R session and launches the celery worker using reticulate
        (see omworker.R)

    Notebooks (jobs):

        R Jupyter notebooks work the very same way as Python notebooks, except
        that they specify the R kernel. From an omega-ml perspective, there is
        no difference in starting a Jupyter notebook with any kernel.

        To initialize the omegaml session within the notebook add a first cell
        that loads omegaml into the R environment started by the R kernel:

            # first cell of notebook
            library(reticulate)
            om <- import("omegaml")
            om$runtimes$rsystem$load()

Usage:

    # R
    library(reticulate)
    om <- import("omegaml")
    om$runtimes$rsystem$load()

    # operations
    om$datasets$list()
    om$datasets$put(df, 'myname')
    df <- om$datasets$get('myname')

    etc.

See Also

    https://github.com/rstudio/reticulate
"""
import shutil

from pathlib import Path

import sys

# R:reticulate injects the r helper object into the main module
# there is not much documentation on this, however a few hints
# https://rstudio.github.io/reticulate/news/index.html#reticulate-111
# https://github.com/rstudio/reticulate/issues/235#issuecomment-388150587
# https://github.com/rstudio/reticulate/blob/4967abddb35865a1be6ad839298695f1481bcade/R/python.R#L1290
# keep track whether R has already sourced omegaml helpers
R_LOAD_STATUS = {}


def rhelper(init=True):
    r_session = getattr(sys.modules['__main__'], 'r', None)
    load(r_session) if r_session and init and id(r_session) not in R_LOAD_STATUS else None
    return r_session


def load(r_session=None):
    """ load omegamlr helper methods

    This is the equivalent of a future library(omegamlr) statement.
    It loads all dependencies and functions required for omegaml R support.
    """
    r = r_session or rhelper(init=False)
    omegamlr_path = Path(__file__).parent / 'omegamlr.R'
    r.source(str(omegamlr_path))
    R_LOAD_STATUS[id(r)] = True


def start_worker(om, queue=None):
    """ start the omega runtime worker from within R env

    This starts the omega worker within a R session, ensuring
    the R helper object is initialized. This is the equivalent
    of '$ om runtime celery rworker'

    Usage:
        # omworker.R
        library(reticulate)
        om <- import("omegaml")
        om$runtimes$rsystem$start_worker(om)

    Args:
        om (Omega): the omega instance

    Returns:
        this is a blocking call, does not stop until worker is stopped
    """
    argv = 'worker --loglevel=DEBUG -E'.split(' ')
    if queue:
        argv.extend(f'-Q {queue}'.split(' '))
    om.runtime.celeryapp.worker_main(argv=argv)

#: determine if R is installed
r_available = shutil.which('Rscript') is not None
#: determine if we are running inside R reticulate
inside_r = rhelper() is not None



"""
omega runtime job tasks
"""
from __future__ import absolute_import

import datetime
from celery import shared_task
from celery.utils.log import get_task_logger

from omegaml.celery_util import OmegamlTask, sanitized
from omegaml.documents import MDREGISTRY


class NotebookTask(OmegamlTask):
    abstract = True

    def on_success(self, retval, task_id, *args, **kwargs):
        super().on_success(retval, task_id, *args, **kwargs)
        try:
            om = self.om
            args, kwargs = args[0:2]
            nbfile = args[0]
            meta = om.jobs.metadata(nbfile)
            attrs = meta.attributes
            attrs['state'] = 'SUCCESS'
            attrs['task_id'] = task_id
            meta.kind = MDREGISTRY.OMEGAML_JOBS

            if not kwargs:
                pass
            else:
                attrs['last_run_time'] = kwargs.get('run_at')
                attrs['next_run_time'] = kwargs.get('next_run_time')

            meta.attributes = attrs
            meta.save()
        except:
            pass

    def on_failure(self, retval, task_id, *args, **kwargs):
        super().on_failure(retval, task_id, *args, **kwargs)
        try:
            om = self.om
            args, kwargs = args[0:2]
            nbfile = args[0]
            meta = om.jobs.metadata(nbfile)
            attrs = meta.attributes
            attrs['state'] = 'FAILURE'
            attrs['task_id'] = task_id
            meta.kind = MDREGISTRY.OMEGAML_JOBS

            if not kwargs:
                pass
            else:
                attrs['last_run_time'] = kwargs.get('run_at')
                attrs['next_run_time'] = kwargs.get('next_run_time')

            meta.attributes = attrs
            meta.save()
        except:
            pass


@shared_task(bind=True, base=NotebookTask)
def run_omegaml_job(self, nb_file, event=None, timeout=None, **kwargs):
    """
    runs omegaml job
    """
    result = self.om.jobs.run_notebook(nb_file, event=event, timeout=timeout)
    return sanitized(result)


@shared_task(base=NotebookTask, bind=True)
def schedule_omegaml_job(self, nb_file, **kwargs):
    """
    schedules the running of omegaml job
    """
    result = self.om.jobs.schedule(nb_file, run_at=kwargs.get('run_at'))
    return sanitized(result)


@shared_task(base=OmegamlTask, bind=True)
def execute_scripts(self, **kwargs):
    """
    run scheduled jobs
    """
    logger = get_task_logger(self.name)
    om = self.om
    now = kwargs.get('now') or datetime.datetime.now()
    # get pending tasks, execute if time is right
    for job_meta in om.jobs.list(raw=True):
        if job_meta.name.startswith('results'):
            # ignore any scheduled results
            continue
        logger.debug("***** {}".format(job_meta))
        triggers = job_meta.attributes.get('triggers', [])
        # run pending jobs
        pending = (trigger for trigger in triggers
                   if trigger['event-kind'] == 'scheduled' and trigger['status'] == 'PENDING')
        for trigger in pending:
            run_at = trigger['run-at']
            logger.info("***** now={} run_at={}".format(now, run_at))
            if now >= run_at:
                om.runtime.job(job_meta.name).run(event=trigger['event'])
                # immediately schedule for next time
                om.jobs.schedule(job_meta.name, last_run=now)



"""
omegaml write and read large CSV files from/to remote or local file paths
Supports s3, hdfs, http/s, sftp, scp, ssh

Usage:
    # read
    om.read_csv('protocol://path/test.csv', 'foobar')

    # write
    mdf = om.datasets.getl('tripdata')
    mdf.to_csv('protocol://path/test.csv')

    -- note path can be local, s3, http/s, scp, ftp, ssh (anything that smart_open supports,
       see https://github.com/RaRe-Technologies/smart_open

(c) 2019, 2020 omegaml.io by oneseven GmbH, Zurich, Switzerland
"""
import pandas as pd
from tqdm import tqdm

try:
    from smart_open import open as open_file
except:
    open_file = open


class IOToolsMDFMixin:
    def to_csv(mdf, csvfn, chunksize=10000, maxobs=None, apply=None, mode='w', open_kwargs=None, **kwargs):
        """
        write MDataframe to s3, hdfs, http/s, sftp, scp, ssh, write to om.datasets

        Usage:
            mdf = om.datasets.getl('name')
            mdf.to_csv('/path/filename.csv')
        """
        df_iter = mdf.iterchunks(chunksize=chunksize)
        _chunked_to_csv(df_iter, csvfn, mode, apply, open_kwargs=open_kwargs, **kwargs)


class IOToolsStoreMixin:
    def read_csv(self, csvfn, name, chunksize=10000, append=False, apply=None, mode='r',
                 open_kwargs=None, **kwargs):
        """
        read large files from s3, hdfs, http/s, sftp, scp, ssh, write to om.datasets

        Usage:

            To insert a local csv file into a dataset::

                om.datasets.read_csv('/path/filename.csv', 'dataset-name')

            To insert a file stored in any of the supported locations specify
            its fully qualified location and filename. The specific format must
            be specified according to the `smart_open`_ library::

                om.datasets.read_csv('https://...', 'dataset-name')
                om.datasets.read_csv('s3://...', 'dataset-name')
                om.datasets.read_csv('hdfs://...', 'dataset-name')
                om.datasets.read_csv('sftp://...', 'dataset-name')
                om.datasets.read_csv('scp://...', 'dataset-name')
                om.datasets.read_csv('ssh://...', 'dataset-name')

            Optionally define a function to receives each chunk as a dataframe
            and apply further processing (e.g. transformations, filtering)::

                def process(df):
                    # apply any processing to df
                    return df

                om.datasets.read_csv(...., apply=process)

        Args:
            csvfn (str): the fully qualified path and name of the csv file,
               according to the `smart_open`_ library
            chunksize (int): the size of each chunk processed before writing
                to the dataset
            append (bool): if True, appends to the dataset. defaults to False
            apply (callable): if specified, each chunk is forwarded as a
               DataFrame and the returned result is inserted to the dataset.
               Use this for transformations or filtering
            mode (str): file open mode, defaults to r
            open_kwargs (dict): additional kwargs to `smart_open`_
            **kwargs: additional kwargs are passed to ``pandas.read_csv``

        Returns:
            MDataFrame

        See Also:

            * `smart_open` https://github.com/RaRe-Technologies/smart_open
            * `pandas.read_csv`

        .. _smart_open: https://github.com/RaRe-Technologies/smart_open
        """
        store = self
        open_kwargs = open_kwargs or {}
        with open_file(csvfn, mode=mode, **open_kwargs) as fin:
            it = pd.read_csv(fin, chunksize=chunksize, iterator=True, **kwargs)
            pbar = tqdm(it)
            try:
                for i, chunkdf in enumerate(pbar):
                    if apply:
                        result = apply(chunkdf)
                        chunkdf = chunkdf if result is None else result
                    store.put(chunkdf, name, append=(i > 0) or append)
            finally:
                pbar.close()
        return store.getl(name)

    def to_csv(self, name, csvfn, chunksize=10000, apply=None, mode='w', open_kwargs=None, **kwargs):
        """ write any dataframe to s3, hdfs, http/s, sftp, scp, ssh

        Usage:

            To write a dataframe::

                om.datasets.write_csv('dataframe-dataset', '/path/to/filename')

            To write a large dataframe in chunks:

                om.datasets.write_csv('dataframe-dataset', '/path/to/filename',
                chunksize=100)

        Args:
            name (str): the name of the dataframe dataset
            csvfn (str): the fully qualified path and name of the csv file,
               according to the `smart_open`_ library
            chunksize (int): the size of each chunk processed before writing
                to the file
            apply (callable): if specified, each chunk is forwarded as a
               DataFrame and the returned result is written to the file.
               Use this for transformations or filtering
            mode (str): file open mode, defaults to w
            open_kwargs (dict): additional kwargs to `smart_open`_
            **kwargs: additional kwargs are passed to ``pandas.to_csv``

        See Also:

            * ``pandas.to_csv``
        """
        df_iter = self.get(name, chunksize=chunksize)
        _chunked_to_csv(df_iter, csvfn, mode, apply, open_kwargs=open_kwargs, **kwargs)


def _chunked_to_csv(df_iter, csvfn, mode, apply, open_kwargs=None, **kwargs):
    open_kwargs = open_kwargs or {}
    with open_file(csvfn, mode, **open_kwargs) as fout:
        for i, chunkdf in tqdm(enumerate(df_iter)):
            if apply:
                result = apply(chunkdf)
                chunkdf = chunkdf if result is None else result
            chunkdf.to_csv(fout, **kwargs)



from __future__ import absolute_import

import logging

from omegaml.util import extend_instance

logger = logging.getLogger(__file__)


class OmegaJobProxy(object):
    """
    proxy to a remote job in a celery worker

    Usage:

        .. code::

            om = Omega()
            # result is AsyncResult, use .get() to return it's result
            result = om.runtime.job('foojob').run()
            result.get()

            # result is AsyncResult, use .get() to return it's result
            result = om.runtime.job('foojob').schedule()
            result.get()
    """

    def __init__(self, jobname, runtime=None):
        self.jobname = jobname
        self.runtime = runtime
        self._apply_mixins()

    def run(self, timeout=None, **kwargs):
        """
        submit the job for immediate execution

        Args:
            timeout (int): optional, timeout in seconds
            **kwargs: kwargs to CeleryTask.delay

        Returns:
            AsyncResult
        """
        job_run = self.runtime.task('omegaml.notebook.tasks.run_omegaml_job')
        return job_run.delay(self.jobname, timeout=timeout, **kwargs)

    def schedule(self, **kwargs):
        """
        schedule the job for repeated execution

        Args:
            **kwargs: see OmegaJob.schedule()
        """
        job_run = self.runtime.task('omegaml.notebook.tasks.schedule_omegaml_job')
        return job_run.delay(self.jobname, **kwargs)

    def _apply_mixins(self):
        """
        apply mixins in defaults.OMEGA_STORE_MIXINS
        """
        for mixin in self.runtime.omega.defaults.OMEGA_JOBPROXY_MIXINS:
            extend_instance(self, mixin)



from collections import defaultdict
from copy import deepcopy
from hashlib import sha1

# byte string
_u8 = lambda t: t.encode('UTF-8', 'replace') if isinstance(t, str) else t


class ModelVersionMixin(object):
    """
    Versioning support for models

    Usage:
        # create a new version
        om.models.put(model, 'mymodel')

        # get the most recent version
        om.models.get('mymodel', version=-1)

        # tag a model, get specific tag
        om.models.put(model, 'mymodel', tag='foo')
        om.models.get('mymodel', tag='foo')

        # get a specific tag included in name
        om.models.get('mymodel@foo')

        # specify the commit id yourself (e.g. when integrating with git)
        # note this does not interact with git in any way
        om.models.put(model, 'mymodel', commit=<git ref>'

    Notes:
        * every put will create a new version
        * it is not possible to delete a version
        * the versioning is purely based on model Metadata,
          all functionality will continue to work as before
    """

    def put(self, obj, name, tag=None, commit=None, previous='latest', noversion=False, **kwargs):
        # create a new version
        base_name, tag, version = self._base_name(name, tag)
        meta = super().put(obj, base_name, **kwargs)
        if self._model_version_applies(name) and not noversion:
            self._ensure_versioned(meta)
            meta = self._put_version(obj, meta, tag=tag, commit=commit, previous=previous, **kwargs)
        return meta

    def get(self, name, commit=None, tag=None, version=-1, **kwargs):
        if not self._model_version_applies(name):
            return super().get(name, **kwargs)
        meta = self.metadata(name, commit=commit, tag=tag, version=version)
        actual_name = name
        if meta:
            self._ensure_versioned(meta)
            actual_name = self._model_version_actual_name(name, tag=tag,
                                                          commit=commit,
                                                          version=version)
        return super().get(actual_name, **kwargs)

    def drop(self, name, force=False, version=-1, commit=None, tag=None, **kwargs):
        # TODO implement drop to support deletion of specific versions
        if False and self._model_version_applies(name):
            # this messes up the version history of the base object!
            name = self._model_version_actual_name(name, tag=tag, commit=commit, version=version)
        return super().drop(name, force=force, **kwargs)

    def metadata(self, name, bucket=None, prefix=None, version=None, commit=None, tag=None, raw=False, **kwargs):
        if not self._model_version_applies(name):
            return super().metadata(name)
        base_meta, base_commit, base_version = self._base_metadata(name, bucket=bucket, prefix=prefix)
        commit = commit or base_commit
        version = base_version or version or -1
        if raw and base_meta and 'versions' in base_meta.attributes:
            # the actual version's metadata is requested
            actual_name = self._model_version_actual_name(name, tag=tag,
                                                          commit=commit,
                                                          version=version,
                                                          bucket=bucket,
                                                          prefix=prefix)
            meta = super().metadata(actual_name, bucket=bucket, prefix=prefix)
        elif base_meta:
            # there is a versioned entry, return the base
            meta = base_meta
        else:
            # no version found, return the actual object
            meta = super().metadata(name)
        return meta

    def revisions(self, name):
        if self._model_version_applies(name):
            meta = self.metadata(name)
            versions = meta.attributes.get('versions', {})
            commits = versions.get('commits', [])
            tags = versions.get('tags', {})
            commit_tags = defaultdict(list)
            for k, v in tags.items():
                commit_tags[v].append(k)
            revisions = [
                (commit['ref'], commit_tags.get(commit['ref'], ''))
                for commit in commits
            ]
            return revisions
        raise NotImplementedError

    def _base_metadata(self, name, **kwargs):
        name, tag, version = self._base_name(name)
        meta = super().metadata(name, **kwargs)
        return meta, tag, version

    def _base_name(self, name, tag=None, **kwargs):
        # return actual name without version tag
        version = None
        if '^' in str(name):
            version = -1 * (name.count('^') + 1)
            name = name.split('^')[0].split('@')[0]
        elif '@' in name:
            name, tag = name.split('@')
        return name, tag, version

    def _model_version_actual_name(self, name, tag=None, commit=None,
                                   version=None, bucket=None, prefix=None):
        meta, name_tag, name_version = self._base_metadata(name, bucket=bucket, prefix=prefix)
        tag = tag or name_tag
        commit = commit or tag
        version = name_version or version or -1
        if meta is not None and 'versions' in meta.attributes:
            actual_name = meta.name
            # we have an existing versioned object
            if tag or commit:
                if tag and tag in meta.attributes['versions']['tags']:
                    version_hash = meta.attributes['versions']['tags'][tag]
                    actual_name = self._model_version_store_key(meta.name, version_hash)
                elif commit:
                    actual_name = self._model_version_store_key(meta.name, commit)
                else:
                    actual_name = name
            else:
                if version == -1 or abs(version) <= len(meta.attributes['versions']['commits']):
                    actual_name = meta.attributes['versions']['commits'][version]['name']
        else:
            actual_name = name
        return actual_name

    def _model_version_hash(self, meta):
        hasher = sha1()
        hasher.update(_u8(meta.name))
        hasher.update(_u8(str(meta.modified)))
        return hasher.hexdigest()

    def _model_version_store_key(self, name, version_hash):
        return '_versions/{}/{}'.format(name, version_hash)

    def _model_version_applies(self, name):
        return self.prefix.startswith('models/')

    def _ensure_versioned(self, meta):
        if 'versions' not in meta.attributes:
            meta.attributes['versions'] = {}
            meta.attributes['versions']['tags'] = {}
            meta.attributes['versions']['commits'] = []
            meta.attributes['versions']['tree'] = {}

    def _put_version(self, obj, meta, tag=None, commit=None, previous=None, **kwargs):
        version_hash = commit or self._model_version_hash(meta)
        previous = meta.attributes['versions']['tags'].get(previous) or previous
        version_name = self._model_version_store_key(meta.name, version_hash)
        version_meta = self.put(obj, version_name, noversion=True, **kwargs)
        version_meta.attributes = deepcopy(meta.attributes)
        version_meta.attributes.update(kwargs.get('attributes', {}))
        if 'versions' in version_meta.attributes:
            del version_meta.attributes['versions']
        version_meta.save()
        meta.attributes['versions']['commits'].append(dict(name=version_meta.name, ref=version_hash))
        meta.attributes['versions']['tree'][version_hash] = previous
        meta.attributes['versions']['tags']['latest'] = version_hash
        if tag:
            meta.attributes['versions']['tags'][tag] = version_hash
        return meta.save()



from __future__ import absolute_import

import numpy as np
import pandas as pd
from bson import Code
from numpy import isscalar
from pymongo.collection import Collection
from uuid import uuid4

from omegaml.store import qops
from omegaml.store.filtered import FilteredCollection
from omegaml.store.query import Filter, MongoQ
from omegaml.store.queryops import MongoQueryOps
from omegaml.util import make_tuple, make_list, restore_index, \
    cursor_to_dataframe, restore_index_columns_order, PickableCollection, extend_instance, json_normalize, ensure_index

INSPECT_CACHE = []


class MGrouper(object):
    """
    a Grouper for MDataFrames
    """
    STATS_MAP = {
        'std': 'stdDevSamp',
        'mean': 'avg',
    }

    def __init__(self, mdataframe, collection, columns, sort=True):
        self.mdataframe = mdataframe
        self.collection = collection
        self.columns = make_tuple(columns)
        self.should_sort = sort

    def __getattr__(self, attr):
        if attr in self.columns:
            return MSeriesGroupby(self, self.collection, attr)

        def statfunc():
            columns = self.columns or self._non_group_columns()
            return self.agg({col: attr for col in columns})

        return statfunc

    def __getitem__(self, item):
        return self.__getattr__(item)

    def agg(self, specs):
        """
        shortcut for .aggregate
        """
        return self.aggregate(specs)

    def aggregate(self, specs, **kwargs):
        """
        aggregate by given specs

        See the following link for a list of supported operations.
        https://docs.mongodb.com/manual/reference/operator/aggregation/group/

        :param specs: a dictionary of { column : function | list[functions] }
           pairs.
        """

        def add_stats(specs, column, stat):
            specs['%s_%s' % (column, stat)] = {
                '$%s' % MGrouper.STATS_MAP.get(stat, stat): '$%s' % column}

        # generate $group command
        _specs = {}
        for column, stats in specs.items():
            stats = make_tuple(stats)
            for stat in stats:
                add_stats(_specs, column, stat)
        groupby = qops.GROUP(columns=self.columns,
                             **_specs)
        # execute and return a dataframe
        pipeline = self._amend_pipeline([groupby])
        data = self.collection.aggregate(pipeline, allowDiskUse=True)

        def get_data():
            # we need this to build a pipeline for from_records
            # to process, otherwise the cursor will be exhausted already
            for group in data:
                _id = group.pop('_id')
                if isinstance(_id, dict):
                    group.update(_id)
                yield group

        df = pd.DataFrame.from_records(get_data())
        columns = make_list(self.columns)
        if columns:
            df = df.set_index(columns, drop=True)
        return df

    def _amend_pipeline(self, pipeline):
        """ amend pipeline with default ops on coll.aggregate() calls """
        if self.should_sort:
            sort = qops.SORT(**dict(qops.make_sortkey('_id')))
            pipeline.append(sort)
        return pipeline

    def _non_group_columns(self):
        """ get all columns in mdataframe that is not in columns """
        return [col for col in self.mdataframe.columns
                if col not in self.columns and col != '_id'
                and not col.startswith('_idx')
                and not col.startswith('_om#')]

    def _count(self):
        count_columns = self._non_group_columns()
        if len(count_columns) == 0:
            count_columns.append('_'.join(self.columns) + '_count')
        groupby = {
            "$group": {
                "_id": {k: "$%s" % k for k in self.columns},
            }
        }
        for k in count_columns:
            groupby['$group']['%s' % k] = {"$sum": 1}
        pipeline = self._amend_pipeline([groupby])
        if self.should_sort:
            sort = qops.SORT(**dict(qops.make_sortkey('_id')))
            pipeline.append(sort)
        return list(self.collection.aggregate(pipeline, allowDiskUse=True))

    def count(self):
        """ return counts by group columns """
        counts = self._count()
        # remove mongo object _id
        for group in counts:
            group.update(group.pop('_id'))
        # transform results to dataframe, then return as pandas would
        resultdf = pd.DataFrame(counts).set_index(make_list(self.columns),
                                                  drop=True)
        return resultdf

    def __iter__(self):
        """ for each group returns the key and a Filter object"""
        # reduce count to only one column
        groups = getattr(self, self.columns[0])._count()
        for group in groups:
            keys = group.get('_id')
            data = self.mdataframe._clone(query=keys)
            yield keys, data


class MLocIndexer(object):
    """
    implements the LocIndexer for MDataFrames
    """

    def __init__(self, mdataframe, positional=False):
        self.mdataframe = mdataframe
        # if positional, any loc[spec] will be applied on the rowid only
        self.positional = positional
        # indicator will be set true if loc specs are from a range type (list, tuple, np.ndarray)
        self._from_range = False

    def __getitem__(self, specs):
        """
        access by index

        use as mdf.loc[specs] where specs is any of

        * a list or tuple of scalar index values, e.g. .loc[(1,2,3)]
        * a slice of values e.g. .loc[1:5]
        * a list of slices, e.g. .loc[1:5, 2:3]

        :return: the sliced part of the MDataFrame
        """
        filterq, projection = self._get_filter(specs)
        df = self.mdataframe
        if filterq:
            df = self.mdataframe.query(filterq)
            df.from_loc_indexer = True
            df.from_loc_range = self._from_range
        if projection:
            df = df[projection]
        if isinstance(self.mdataframe, MSeries):
            df = df._as_mseries(df.columns[0])
        if getattr(df, 'immediate_loc', False):
            df = df.value
        return df

    def __setitem__(self, specs, value):
        raise NotImplementedError()

    def _get_filter(self, specs):
        filterq = []
        projection = None
        if self.positional:
            idx_cols = ['_om#rowid']
        else:
            idx_cols = self.mdataframe._get_frame_index()
        flt_kwargs = {}
        enumerable_types = (list, tuple, np.ndarray)
        if isinstance(specs, np.ndarray):
            specs = specs.tolist()
        if (isinstance(specs, enumerable_types)
              and isscalar(specs[0]) and len(idx_cols) == 1
              and not any(isinstance(s, slice) for s in specs)):
            # single column index with list of scalar values
            if (self.positional and isinstance(specs, tuple) and len(specs) == 2
                  and all(isscalar(v) for v in specs)):
                # iloc[int, int] is a cell access
                flt_kwargs[idx_cols[0]] = specs[0]
                projection = self._get_projection(specs[1])
            else:
                flt_kwargs['{}__in'.format(idx_cols[0])] = specs
                self._from_range = True
        elif isinstance(specs, (int, str)):
            flt_kwargs[idx_cols[0]] = specs
        else:
            specs = make_tuple(specs)
            # list/tuple of slices or scalar values, or MultiIndex
            for i, spec in enumerate(specs):
                if i < len(idx_cols):
                    col = idx_cols[i]
                    if isinstance(spec, slice):
                        self._from_range = True
                        start, stop = spec.start, spec.stop
                        if start is not None:
                            flt_kwargs['{}__gte'.format(col)] = start
                        if stop is not None:
                            if isinstance(stop, int):
                                stop -= int(self.positional)
                            flt_kwargs['{}__lte'.format(col)] = stop
                    elif isinstance(spec, enumerable_types) and isscalar(spec[0]):
                        self._from_range = True
                        # single column index with list of scalar values
                        # -- convert to list for PyMongo serialization
                        if isinstance(spec, np.ndarray):
                            spec = spec.tolist()
                        flt_kwargs['{}__in'.format(col)] = spec
                    elif isscalar(col):
                        flt_kwargs[col] = spec
                else:
                    # we're out of index columns, let's look at columns
                    cur_proj = self._get_projection(spec)
                    # if we get a single column, that's the projection
                    # if we had a single column, now need to add more, create a list
                    # if we have a list already, extend that
                    if projection is None:
                        projection = cur_proj
                    elif isinstance(projection, list):
                        projection.extend(cur_proj)
                    else:
                        projection = [projection, cur_proj]
        if flt_kwargs:
            filterq.append(MongoQ(**flt_kwargs))
        finalq = None
        for q in filterq:
            if finalq:
                finalq |= q
            else:
                finalq = q
        return finalq, projection

    def _get_projection(self, spec):
        columns = self.mdataframe.columns
        if np.isscalar(spec):
            return [spec]
        if isinstance(spec, (tuple, list)):
            assert all(columns.index(col) for col in columns)
            return spec
        if isinstance(spec, slice):
            start, stop = spec.start, spec.stop
            if all(isinstance(v, int) for v in (start, stop)):
                start, stop, step = spec.indices(len(columns))
            else:
                start = columns.index(start) if start is not None else 0
                stop = columns.index(stop) + 1 if stop is not None else len(columns)
            return columns[slice(start, stop)]
        raise IndexError


class MPosIndexer(MLocIndexer):
    """
    implements the position-based indexer for MDataFrames
    """

    def __init__(self, mdataframe):
        super(MPosIndexer, self).__init__(mdataframe, positional=True)

    def _get_projection(self, spec):
        columns = self.mdataframe.columns
        if np.isscalar(spec):
            return columns[spec]
        if isinstance(spec, (tuple, list)):
            return [col for i, col in enumerate(spec) if i in spec]
        if isinstance(spec, slice):
            start, stop = slice.start, slice.stop
            if start and not isinstance(start, int):
                start = 0
            if stop and not isinstance(stop, int):
                # sliced ranges are inclusive
                stop = len(columns)
            return columns[slice(start, stop)]
        raise IndexError


class MSeriesGroupby(MGrouper):
    """
    like a MGrouper but limited to one column
    """

    def count(self):
        """
        return series count

        :return: counts by group
        """
        # MGrouper will insert a _count column, see _count(). we remove
        # that column again and return a series named as the group column
        resultdf = super(MSeriesGroupby, self).count()
        count_column = [col for col in resultdf.columns
                        if col.endswith('_count')][0]
        new_column = count_column.replace('_count', '')
        resultdf = resultdf.rename(columns={count_column: new_column})
        return resultdf[new_column]


class MDataFrame(object):
    """
    A DataFrame for mongodb

    Performs out-of-core, lazy computOation on a mongodb cluster.
    Behaves like a pandas DataFrame. Actual results are returned
    as pandas DataFrames.
    """

    STATFUNCS = ['mean', 'std', 'min', 'max', 'sum', 'var']

    def __init__(self, collection, columns=None, query=None,
                 limit=None, skip=None, sort_order=None,
                 force_columns=None, immediate_loc=False, auto_inspect=False,
                 normalize=False, raw=False,
                 parser=None,
                 preparefn=None, from_loc_range=False, metadata=None, **kwargs):
        self.collection = PickableCollection(collection)
        # columns in frame
        self.columns = make_tuple(columns) if columns else self._get_fields(raw=raw)
        self.columns = [str(col) for col in self.columns]
        # columns to sort by, defaults to not sorted
        self.sort_order = sort_order
        # top n documents to fetch
        self.head_limit = limit
        # top n documents to skip before returning
        self.skip_topn = skip
        # filter criteria
        self.filter_criteria = query or {}
        # force columns -- on output add columns not present
        self.force_columns = force_columns or []
        # was this created from the loc indexer?
        self.from_loc_indexer = kwargs.get('from_loc_indexer', False)
        # was the loc index used a range? Else a single value
        self.from_loc_range = from_loc_range
        # setup query for filter criteries, if provided
        if self.filter_criteria:
            # make sure we have a filtered collection with the criteria given
            if isinstance(self.filter_criteria, dict):
                self.query_inplace(**self.filter_criteria)
            elif isinstance(self.filter_criteria, Filter):
                self.query_inplace(self.filter_criteria)
            else:
                raise ValueError('Invalid query specification of type {}'.format(type(self.filter_criteria)))
        # if immediate_loc is True, .loc and .iloc always evaluate
        self.immediate_loc = immediate_loc
        # __array__ will return this value if it is set, set it otherwise
        self._evaluated = None
        # set true to automatically capture inspects on .value. retrieve using .inspect(cached=True)
        self.auto_inspect = auto_inspect
        self._inspect_cache = INSPECT_CACHE
        # apply mixins
        self._applyto = str(self.__class__)
        self._apply_mixins()
        # parser to parse documents to dataframe
        self._parser = json_normalize if normalize else parser
        # prepare function to be applied just before returning from .value
        self._preparefn = preparefn
        # keep technical fields like _id, _idx etc
        self._raw = raw
        # metadata stored by omegaml (equiv. of metadata.kind_meta)
        self.metadata = metadata or dict()

    def _apply_mixins(self, *args, **kwargs):
        """
        apply mixins in defaults.OMEGA_MDF_MIXINS
        """
        from omegaml import settings
        defaults = settings()
        for mixin, applyto in defaults.OMEGA_MDF_MIXINS:
            if any(v in self._applyto for v in applyto.split(',')):
                extend_instance(self, mixin, *args, **kwargs)

    def __getstate__(self):
        # pickle support. note that the hard work is done in PickableCollection
        data = dict(self.__dict__)
        data.update(_evaluated=None)
        data.update(_inspect_cache=None)
        data.update(auto_inspect=self.auto_inspect)
        data.update(_preparefn=self._preparefn)
        data.update(_parser=self._parser)
        data.update(_raw=self._raw)
        data.update(collection=self.collection)
        return data

    def __reduce__(self):
        state = self.__getstate__()
        args = self.collection,
        return _mdf_remake, args, state

    def __setstate__(self, state):
        # pickle support. note that the hard work is done in PickableCollection
        for k, v in state.items():
            setattr(self, k, v)

    def _getcopy_kwargs(self, without=None):
        """ return all parameters required on a copy of this MDataFrame """
        kwargs = dict(columns=self.columns,
                      sort_order=self.sort_order,
                      limit=self.head_limit,
                      skip=self.skip_topn,
                      from_loc_indexer=self.from_loc_indexer,
                      from_loc_range=self.from_loc_range,
                      immediate_loc=self.immediate_loc,
                      metadata=self.metadata,
                      query=self.filter_criteria,
                      auto_inspect=self.auto_inspect,
                      preparefn=self._preparefn)
        [kwargs.pop(k) for k in make_tuple(without or [])]
        return kwargs

    def __array__(self, dtype=None):
        # FIXME inefficient. make MDataFrame a drop-in replacement for any numpy ndarray
        # this evaluates every single time
        if self._evaluated is None:
            self._evaluated = array = self.value.values
        else:
            array = self._evaluated
        return array

    def __getattr__(self, attr):
        if attr in MDataFrame.STATFUNCS:
            return self.statfunc(attr)
        if attr in self.columns:
            kwargs = self._getcopy_kwargs()
            kwargs.update(columns=attr)
            return MSeries(self.collection, **kwargs)
        raise AttributeError(attr)

    def __getitem__(self, cols_or_slice):
        """
        select and project by column, columns, slice, masked-style filter

        Masked-style filters work similar to pd.DataFrame/Series masks
        but do not actually return masks but an instance of Filter. A
        Filter is a delayed evaluation on the data frame.

            # select all rows where any column is == 5
            mdf = MDataFrame(coll)
            flt = mdf == 5
            mdf[flt]
            =>

        :param cols_or_slice: single column (str), multi-columns (list),
          slice to select columns or a masked-style
        :return: filtered MDataFrame or MSeries
        """
        if isinstance(cols_or_slice, str):
            # column name => MSeries
            return self._as_mseries(cols_or_slice)
        elif isinstance(cols_or_slice, int):
            # column number => MSeries
            column = self.columns[cols_or_slice]
            return self._as_mseries(column)
        elif isinstance(cols_or_slice, (tuple, list)):
            # list of column names => MDataFrame subset on columns
            kwargs = self._getcopy_kwargs()
            kwargs.update(columns=cols_or_slice)
            return MDataFrame(self.collection, **kwargs)
        elif isinstance(cols_or_slice, Filter):
            kwargs = self._getcopy_kwargs()
            kwargs.update(query=cols_or_slice.query)
            return MDataFrame(self.collection, **kwargs)
        elif isinstance(cols_or_slice, np.ndarray):
            return self.iloc[cols_or_slice]
        raise ValueError('unknown accessor type %s' % type(cols_or_slice))

    def __setitem__(self, column, value):
        # True for any scalar type, numeric, bool, string
        if np.isscalar(value):
            result = self.collection.update_many(filter=self.filter_criteria,
                                                 update=qops.SET(column, value))
            self.columns.append(column)
        return self

    def _clone(self, collection=None, **kwargs):
        # convenience method to clone itself with updates
        collection = collection if collection is not None else self.collection
        return self.__class__(collection, **kwargs,
                              **self._getcopy_kwargs(without=list(kwargs.keys())))

    def statfunc(self, stat):
        aggr = MGrouper(self, self.collection, [], sort=False)
        return getattr(aggr, stat)

    def groupby(self, columns, sort=True):
        """
        Group by a given set of columns

        :param columns: the list of columns
        :param sort: if True sort by group key
        :return: MGrouper
        """
        return MGrouper(self, self.collection, columns, sort=sort)

    def _get_fields(self, raw=False):
        result = []
        doc = self.collection.find_one()
        if doc is not None:
            if raw:
                result = list(doc.keys())
            else:
                result = [str(col) for col in doc.keys()
                          if col != '_id'
                          and not col.startswith('_idx')
                          and not col.startswith('_om#')]
        return result

    def _get_frame_index(self):
        """ return the dataframe's index columns """
        doc = self.collection.find_one()
        if doc is None:
            result = []
        else:
            result = restore_index_columns_order(doc.keys())
        return result

    def _get_frame_om_fields(self):
        """ return the dataframe's omega special fields columns """
        doc = self.collection.find_one()
        if doc is None:
            result = []
        else:
            result = [k for k in list(doc.keys()) if k.startswith('_om#')]
        return result

    def _as_mseries(self, column):
        kwargs = self._getcopy_kwargs()
        kwargs.update(columns=make_tuple(column))
        return MSeries(self.collection, **kwargs)

    def inspect(self, explain=False, cached=False, cursor=None, raw=False):
        """
        inspect this dataframe's actual mongodb query

        :param explain: if True explains access path
        """
        if not cached:
            if isinstance(self.collection, FilteredCollection):
                query = self.collection.query
            else:
                query = '*',
            if explain:
                cursor = cursor or self._get_cursor()
                explain = cursor.explain()
            data = {
                'projection': self.columns,
                'query': query,
                'explain': explain or 'specify explain=True'
            }
        else:
            data = self._inspect_cache
        if not (raw or explain):
            data = pd.DataFrame(json_normalize(data))
        return data

    def count(self):
        """
        projected number of rows when resolving
        """
        nrows = len(self)
        counts = pd.Series({
            col: nrows
            for col in self.columns}, index=self.columns)
        return counts

    def __len__(self):
        """
        the projected number of rows when resolving
        """
        # we reduce to just 1 column to reduce speed
        short = self._clone()[self.columns[0]]
        return sum(1 for d in short._get_cursor())

    @property
    def shape(self):
        """
        return shape of dataframe
        """
        return len(self), len(self.columns)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def value(self):
        """
        resolve the query and return a Pandas DataFrame

        :return: the result of the query as a pandas DataFrame
        """
        cursor = self._get_cursor()
        df = self._get_dataframe_from_cursor(cursor)
        if self.auto_inspect:
            self._inspect_cache.append(self.inspect(explain=True, cursor=cursor, raw=True))
        # this ensures the equiv. of pandas df.loc[n] is a Series
        if self.from_loc_indexer:
            if len(df) == 1 and not self.from_loc_range:
                idx = df.index
                df = df.T
                df = df[df.columns[0]]
                if df.ndim == 1 and len(df) == 1 and not isinstance(idx, pd.MultiIndex):
                    # single row single dimension, numeric index only
                    df = df.iloc[0]
            elif (df.ndim == 1 or df.shape[1] == 1) and not self.from_loc_range:
                df = df[df.columns[0]]
        if self._preparefn:
            df = self._preparefn(df)
        return df

    def reset(self):
        # TODO if head(), tail(), query(), .loc/iloc should return a new MDataFrame instance to avoid having a reset need
        self.head_limit = None
        self.skip_topn = None
        self.filter_criteria = {}
        self.force_columns = []
        self.sort_order = None
        self.from_loc_indexer = False
        return self

    def _get_dataframe_from_cursor(self, cursor):
        """
        from the given cursor return a DataFrame
        """
        df = cursor_to_dataframe(cursor, parser=self._parser)
        df = self._restore_dataframe_proper(df)
        return df

    @property
    def _index_meta(self):
        return self.metadata.get('idx_meta') or dict()

    def _restore_dataframe_proper(self, df):
        df = restore_index(df, self._index_meta)
        if '_id' in df.columns and not self._raw:
            df.drop('_id', axis=1, inplace=True)
        if self.force_columns:
            missing = set(self.force_columns) - set(self.columns)
            for col in missing:
                df[col] = np.NaN
        return df

    def _get_cursor(self):
        projection = make_tuple(self.columns)
        projection += make_tuple(self._get_frame_index())
        if not self.sort_order:
            # implicit sort
            projection += make_tuple(self._get_frame_om_fields())
        cursor = self.collection.find(projection=projection)
        if self.sort_order:
            cursor.sort(qops.make_sortkey(make_tuple(self.sort_order)))
        if self.head_limit:
            cursor.limit(self.head_limit)
        if self.skip_topn:
            cursor.skip(self.skip_topn)
        return cursor

    def sort(self, columns):
        """
        sort by specified columns

        :param columns: str of single column or a list of columns. Sort order
                        is specified as the + (ascending) or - (descending)
                        prefix to the column name. Default sort order is
                        ascending.
        :return: the MDataFrame
        """
        self._evaluated = None
        self.sort_order = make_tuple(columns)
        return self

    def head(self, limit=10):
        """
        return up to limit numbers of rows

        :param limit: the number of rows to return. Defaults to 10
        :return: the MDataFrame
        """
        return self._clone(limit=limit)

    def tail(self, limit=10):
        """
        return up to limit number of rows from last inserted values

        :param limit:
        :return:
        """
        tail_n = self.skip(len(self) - limit)
        return self._clone(skip=tail_n)

    def skip(self, topn):
        """
        skip the topn number of rows

        :param topn: the number of rows to skip.
        :return: the MDataFrame
        """
        return self._clone(skip=topn)

    def merge(self, right, on=None, left_on=None, right_on=None,
              how='inner', target=None, suffixes=('_x', '_y'),
              sort=False, inspect=False, filter=None):
        """
        merge this dataframe with another dataframe. only left outer joins
        are currently supported. the output is saved as a new collection,
        target name (defaults to a generated name if not specified).

        :param right: the other MDataFrame
        :param on: the list of key columns to merge by
        :param left_on: the list of the key columns to merge on this dataframe
        :param right_on: the list of the key columns to merge on the other
           dataframe
        :param how: the method to merge. supported are left, inner, right.
           Defaults to inner
        :param target: the name of the collection to store the merge results
           in. If not provided a temporary name will be created.
        :param suffixes: the suffixes to apply to identical left and right
           columns
        :param sort: if True the merge results will be sorted. If False the
           MongoDB natural order is implied.
        :returns: the MDataFrame to the target MDataFrame
        """
        # validate input
        supported_how = ["left", 'inner', 'right']
        assert how in supported_how, "only %s merges are currently supported" % supported_how
        for key in [on, left_on, right_on]:
            if key:
                assert isinstance(
                    key, str), "only single column merge keys are supported (%s)" % key
        if isinstance(right, (Collection, PickableCollection, FilteredCollection)):
            right = MDataFrame(right)
        assert isinstance(
            right, MDataFrame), "both must be MDataFrames, got right=%" % type(right)
        if how == 'right':
            # A right B == B left A
            return right.merge(self, on=on, left_on=right_on, right_on=left_on,
                               how='left', target=target, suffixes=suffixes)
        # generate lookup parameters
        on = on or '_id'
        right_name = self._get_collection_name_of(right, right)
        target_name = self._get_collection_name_of(
            target, '_temp.merge.%s' % uuid4().hex)
        target_field = (
              "%s_%s" % (right_name.replace('.', '_'), right_on or on))
        """
        TODO enable filter criteria on right dataframe. requires changing LOOKUP syntax from 
             equitly to arbitray match 
        
        if right.filter_criteria:
            right_filter = [qops.MATCH(self._get_filter_criteria(**right.filter_criteria))]
        else:
            right_filter = None
        """
        right_filter = None
        lookup = qops.LOOKUP(right_name,
                             key=on,
                             left_key=left_on,
                             right_key=right_on,
                             target=target_field)
        # unwind merged documents from arrays to top-level document fields
        unwind = qops.UNWIND(target_field, preserve=how != 'inner')
        # get all fields from left, right
        project = {}
        for left_col in self.columns:
            source_left_col = left_col
            if left_col == '_id':
                project[left_col] = 1
                continue
            if left_col.startswith('_idx'):
                continue
            if left_col.startswith('_om#'):
                continue
            if left_col != (on or left_on) and left_col in right.columns:
                left_col = '%s%s' % (left_col, suffixes[0])
            project[left_col] = "$%s" % source_left_col
        for right_col in right.columns:
            if right_col == '_id':
                continue
            if right_col.startswith('_idx'):
                continue
            if right_col.startswith('_om#'):
                continue
            if right_col == (on or right_on) and right_col == (on or left_on):
                # if the merge field is the same in both frames, we already
                # have it from left
                continue
            if right_col in self.columns:
                left_col = '%s%s' % (right_col, suffixes[1])
            else:
                left_col = '%s' % right_col
            project[left_col] = '$%s.%s' % (target_field, right_col)
        expected_columns = list(project.keys())
        if '_id' not in project:
            project['_id'] = 0  # never copy objectids to avoid duplicate keys, unless requested
        project = {"$project": project}
        # store merged documents and return an MDataFrame to it
        out = qops.OUT(target_name)
        pipeline = [lookup, unwind, project]
        if filter:
            query = qops.MATCH(self._get_filter_criteria(**filter))
            pipeline.append(query)
        if sort:
            sort_cols = make_list(on or [left_on, right_on])
            sort_key = qops.make_sortkey(sort_cols)
            sort = qops.SORT(**dict(sort_key))
            pipeline.append(sort)
        pipeline.append(out)
        if inspect:
            result = pipeline
        else:
            result = self.collection.aggregate(pipeline, allowDiskUse=True)
            result = MDataFrame(self.collection.database[target_name],
                                force_columns=expected_columns)
        return result

    def append(self, other):
        if isinstance(other, Collection):
            right = MDataFrame(other)
        assert isinstance(
            other, MDataFrame), "both must be MDataFrames, got other={}".format(type(other))
        outname = self.collection.name
        mrout = {
            'merge': outname,
            'nonAtomic': True,
        }
        mapfn = Code("""
        function() {
           this._id = ObjectId();
           if(this['_om#rowid']) {
              this['_om#rowid'] += %s;
           }
           emit(this._id, this);
        }
        """ % len(self))
        reducefn = Code("""
        function(key, value) {
           return value;
        }
        """)
        finfn = Code("""
        function(key, value) {
           return value;
        }
        """)
        other.collection.map_reduce(mapfn, reducefn, mrout, finalize=finfn, jsMode=True)
        unwind = {
            "$replaceRoot": {
                "newRoot": {
                    "$ifNull": ["$value", "$$CURRENT"],
                }
            }
        }
        output = qops.OUT(outname)
        pipeline = [unwind, output]
        self.collection.aggregate(pipeline, allowDiskUse=True)
        return self

    def _get_collection_name_of(self, some, default=None):
        """
        determine the collection name of the given parameter

        returns the collection name if some is a MDataFrame, a Collection
        or a string_type. Otherwise returns default
        """
        if isinstance(some, MDataFrame):
            name = some.collection.name
        elif isinstance(some, Collection):
            name = some.name
        else:
            name = default
        return name

    def _get_filter_criteria(self, *args, **kwargs):
        """
        return mongo query from filter specs

        this uses a Filter to produce the query from the kwargs.

        :param args: a Q object or logical combination of Q objects
           (optional)
        :param kwargs: all AND filter criteria
        """
        if len(args) > 0:
            q = args[0]
            if isinstance(q, MongoQ):
                filter_criteria = Filter(self.collection, q).query
            elif isinstance(q, Filter):
                filter_criteria = Filter(self.collection, q.q).query
        else:
            filter_criteria = Filter(self.collection, **kwargs).query
        return filter_criteria

    def query_inplace(self, *args, **kwargs):
        """
        filters this MDataFrame and returns it.

        Any subsequent operation on the dataframe will have the filter
        applied. To reset the filter call .reset() without arguments.

        :param args: a Q object or logical combination of Q objects
           (optional)
        :param kwargs: all AND filter criteria
        :return: self
        """
        self._evaluated = None
        self.filter_criteria = self._get_filter_criteria(*args, **kwargs)
        self.collection = FilteredCollection(
            self.collection, query=self.filter_criteria)
        return self

    def query(self, *args, **kwargs):
        """
        return a new MDataFrame with a filter criteria

        Any subsequent operation on the new dataframe will have the filter
        applied. To reset the filter call .reset() without arguments.

        Note: Unlike pandas DataFrames, a filtered MDataFrame operates
        on the same collection as the original DataFrame

        :param args: a Q object or logical combination of Q objects
           (optional)
        :param kwargs: all AND filter criteria
        :return: a new MDataFrame with the filter applied
        """
        effective_filter = dict(self.filter_criteria)
        filter_criteria = self._get_filter_criteria(*args, **kwargs)
        if '$and' in effective_filter:
            effective_filter['$and'].extend(filter_criteria.get('$and'))
        else:
            effective_filter.update(filter_criteria)
        coll = FilteredCollection(self.collection, query=effective_filter)
        return self._clone(collection=coll, query=effective_filter)

    def create_index(self, keys, **kwargs):
        """
        create and index the easy way
        """
        keys, kwargs = MongoQueryOps().make_index(keys)
        result = ensure_index(self.collection, keys, **kwargs)
        return result

    def list_indexes(self):
        """
        list all indices in database
        """
        return cursor_to_dataframe(self.collection.list_indexes())

    def iterchunks(self, chunksize=100):
        """
        return an iterator

        Args:
            chunksize (int): number of rows in each chunk

        Returns:
            a dataframe of max. length chunksize
        """
        chunksize = int(chunksize)
        i = 0
        while True:
            chunkdf = self.skip(i).head(chunksize).value
            if len(chunkdf) == 0:
                break
            i += chunksize
            yield chunkdf

    def itertuples(self, chunksize=1000):
        chunksize = int(chunksize)
        __doc__ = pd.DataFrame.itertuples.__doc__

        for chunkdf in self.iterchunks(chunksize=chunksize):
            for row in chunkdf.iterrows():
                yield row

    def iterrows(self, chunksize=1000):
        chunksize = int(chunksize)
        __doc__ = pd.DataFrame.iterrows.__doc__

        for chunkdf in self.iterchunks(chunksize=chunksize):
            if isinstance(chunkdf, pd.DataFrame):
                for row in chunkdf.iterrows():
                    yield row
            else:
                # Series does not have iterrows
                for i in range(0, len(chunkdf), chunksize):
                    yield chunkdf.iloc[i:i+chunksize]

    def iteritems(self):
        __doc__ = pd.DataFrame.iteritems.__doc__
        return self.items()

    def items(self):
        __doc__ = pd.DataFrame.items.__doc__

        for col in self.columns:
            yield col, self[col].value

    @property
    def loc(self):
        """
        Access by index

        Use as mdf.loc[index_value]

        :return: MLocIndexer
        """
        self._evaluated = None
        indexer = MLocIndexer(self)
        return indexer

    @property
    def iloc(self):
        self._evaluated = None
        indexer = MPosIndexer(self)
        return indexer

    def rows(self, start=None, end=None, chunksize=1000):
        # equivalent to .iloc[start:end].iteritems(),
        start, end, chunksize = (int(v) for v in (start, end, chunksize))
        return self.iloc[slice(start, end)].iterchunks(chunksize)


    def __repr__(self):
        kwargs = ', '.join('{}={}'.format(k, v) for k, v in self._getcopy_kwargs().items())
        return "MDataFrame(collection={collection.name}, {kwargs})".format(collection=self.collection,
                                                                           kwargs=kwargs)


class MSeries(MDataFrame):
    """
    Series implementation for MDataFrames

    behaves like a DataFrame but limited to one column.
    """

    def __init__(self, *args, **kwargs):
        super(MSeries, self).__init__(*args, **kwargs)
        # true if only unique values apply
        self.is_unique = False
        # apply mixins
        self._applyto = str(self.__class__)
        self._apply_mixins(*args, **kwargs)

    def __getitem__(self, cols_or_slice):
        if isinstance(cols_or_slice, Filter):
            return MSeries(self.collection, columns=self.columns,
                           query=cols_or_slice.query)
        return super(MSeries, self).__getitem__(cols_or_slice)

    @property
    def name(self):
        return self.columns[0]

    def unique(self):
        """
        return the unique set of values for the series

        :return: MSeries
        """
        self.is_unique = True
        return self

    def _get_cursor(self):
        if self.is_unique:
            # this way indexes get applied
            cursor = self.collection.distinct(make_tuple(self.columns)[0])
        else:
            cursor = super(MSeries, self)._get_cursor()
        return cursor

    @property
    def value(self):
        """
        return the value of the series

        this is a Series unless unique() was called. If unique()
        only distinct values are returned as an array, matching
        the behavior of a Series

        :return: pandas.Series
        """
        cursor = self._get_cursor()
        column = make_tuple(self.columns)[0]
        if self.is_unique:
            # the .distinct() cursor returns a list of values
            # this is to make sure we return the same thing as pandas
            val = [v for v in cursor]
        else:
            val = self._get_dataframe_from_cursor(cursor)
            val = val[column]
            val.name = self.name
            if len(val) == 1 and self.from_loc_indexer:
                val = val.iloc[0]
        if self.auto_inspect:
            self._inspect_cache.append(self.inspect(explain=True, cursor=cursor, raw=True))
        if self._preparefn:
            df = self._preparefn(val)
        return val

    def __repr__(self):
        kwargs = ', '.join('{}={}'.format(k, v) for k, v in self._getcopy_kwargs().items())
        return "MSeries(collection={collection.name}, {kwargs})".format(collection=self.collection,
                                                                        kwargs=kwargs)

    @property
    def shape(self):
        return len(self),


def _mdf_remake(collection):
    # recreate a pickled MDF
    mdf = MDataFrame(collection)
    return mdf



import apispec
import json
import yaml
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin

from omegaml.runtimes.mixins.swagger.helpers import SpecFromModelHelper, SpecFromDatasetHelper, SpecFromScriptHelper, \
    SpecFromServiceHelper


class SwaggerGenerator:
    def __init__(self, om):
        self.om = om
        self.spec = self._make_spec()

    def _make_spec(self, title=None, version='1.0.0'):
        title = title or 'omega-ml service'
        # Create an APISpec
        spec = APISpec(
            title=title,
            version=version,
            openapi_version="2.0",
            plugins=[MarshmallowPlugin()],
        )
        return spec

    def _resource_helper(self, resource, as_service=False):
        RESOURCE_SPEC_HELPERS = {
            'model': SpecFromModelHelper,
            'dataset': SpecFromDatasetHelper,
            'script': SpecFromScriptHelper,
            'service': SpecFromServiceHelper,
        }
        handler = resource if not as_service else 'service'
        return RESOURCE_SPEC_HELPERS[handler]

    def to_json(self, indent=2, stream=None):
        dump = json.dump if stream else json.dumps
        return dump(self.spec.to_dict(), indent=indent)

    def to_yaml(self, stream=None):
        return yaml.dump(self.spec.to_dict(), stream=stream)

    def to_dict(self, **kwargs):
        return self.spec.to_dict()

    def include(self, path, as_service=False):
        for obj in self.om.list(path.replace('datasets/', 'data/')):
            self._include_obj(obj.replace('data/', 'datasets/'), as_service=as_service)

    def validate(self):
        apispec.utils.validate_spec(self.spec)

    def _include_obj(self, path, as_service=False):
        prefix, name = path.split('/', 1)
        store = getattr(self.om, prefix)
        meta = store.metadata(name)
        # cut the s (models => model, datasets => dataset)
        resource = prefix[:-1]
        signature = meta.attributes.get('signature', {})
        spec = self.spec
        resource_helper = self._resource_helper(resource, as_service)
        resource_helper(meta, name, resource, signature, spec, store).render()

    @staticmethod
    def build_swagger(self, include='*', format='yaml', file=None, as_service=False):
        gen = SwaggerGenerator(self.omega)
        includes = include.split(',') if isinstance(include, str) else include
        for path in includes:
            gen.include(path, as_service=as_service)
        formatter = {
            'yaml': gen.to_yaml,
            'dict': gen.to_dict,
            'json': gen.to_json,
        }
        return formatter.get(format)(stream=file)




from os.path import dirname, basename

import os

from omegaml.backends.basedata import BaseDataBackend

try:
    from smart_open import open
except:
    pass


class PythonRawFileBackend(BaseDataBackend):
    """
    OmegaStore backend to support arbitrary files
    """
    KIND = 'python.file'

    @classmethod
    def supports(self, obj, name, open_kwargs=None, **kwargs):
        is_filelike = hasattr(obj, 'read')
        open_kwargs = dict(open_kwargs or {})
        if kwargs.get('kind') == self.KIND:
            is_filelike = self._is_openable(self, obj, **open_kwargs)
        return is_filelike or self._is_path(self, obj)

    def _is_openable(self, obj, **kwargs):
        if 'mode' not in 'kwargs':
            kwargs['mode'] = 'rb'
        try:
            with open(obj, **kwargs) as fin:
                fin.read(1)
        except:
            return False
        return True

    def get(self, name, local=None, mode='wb', open_kwargs=None, **kwargs):
        """
        get a stored file as a file-like object with binary contents or a local file

        Args:
            name (str): the name of the file
            local (str): if set the local path will be created and the file
               stored there. If local does not have an extension it is assumed
               to be a directory name, in which case the file is stored as the
               same name.
            mode (str): the mode to use on .open() for the local file
            open_kwargs (dict): the kwargs to use .open() for the local file
            **kwargs: any kwargs passed to datasets.metadata()

        Returns:
            the file-like output handler (local is None)
            the path to the local file (local is given)

        See also:
            https://docs.python.org/3/glossary.html#term-file-object
            https://docs.python.org/3/glossary.html#term-binary-file
        """
        outf = self.data_store.metadata(name, **kwargs).gridfile
        if local:
            is_filename = '.' in basename(local)
            target_dir = dirname(local) if is_filename else local
            local = local if is_filename else '{local}/{name}'.format(**locals())
            os.makedirs(target_dir, exist_ok=True)
            open_kwargs = open_kwargs or {}
            with open(local, mode=mode, **open_kwargs) as flocal:
                flocal.write(outf.read())
            return local
        return filelike(outf)

    def put(self, obj, name, attributes=None, encoding=None, **kwargs):
        self.data_store.drop(name, force=True)
        storekey = self.data_store.object_store_key(name, 'file', hashed=True)
        gridfile = self._store_to_file(self.data_store, obj, storekey, encoding=encoding)
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()


def filelike(obj):
    actual = obj.get()
    __doc__ = actual.__doc__
    return actual



import shutil

from pathlib import Path

import os
import unittest
import warnings
from shutil import rmtree
from unittest import TestCase

import pandas as pd
from numpy.testing import assert_array_equal
from sklearn.linear_model import LinearRegression

import omegaml.defaults
from omegaml import Omega
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import reshaped, module_available

try:
    from omegaml.backends.mlflow.models import MLFlowModelBackend
    from omegaml.backends.mlflow.registrymodels import MLFlowRegistryBackend
except:
    warnings.warn("mlflow is not installed")
else:
    from mlflow.exceptions import MlflowException
    import mlflow


    @unittest.skipUnless(module_available('mlflow'), 'mlflow not available')
    class TestMLFlowModels(OmegaTestMixin, TestCase):
        @classmethod
        def setUpClass(cls):
            # we do this here to avoid interfering with mlflow dbs during tests
            # due to mlflow's implicit once-only db setup at the first session start
            cls.mlflow_tracking_db = 'sqlite:////tmp/mlflow-t.sqlite'
            cls.mlflow_registry_db = 'sqlite:////tmp/mlflow-r.sqlite'
            cls._clean_mlruns()
            cls._clean_mlflowdbs()

        def setUp(self):
            om = self.om = Omega()
            self.clean()
            om.models.register_backend(MLFlowModelBackend.KIND, MLFlowModelBackend)
            om.models.register_backend(MLFlowRegistryBackend.KIND, MLFlowRegistryBackend)

        def tearDown(self):
            self._enable_mlflow_exit_handling()

        def test_save_mlflow_saved_model_path(self):
            """ test deploying a model saved by MLflow, from path """
            import mlflow

            model_path = os.path.join(omegaml.defaults.OMEGA_TMP, 'mymodel')
            model = LinearRegression()
            X = pd.Series(range(0, 10))
            Y = pd.Series(X) * 2 + 3
            model.fit(reshaped(X), reshaped(Y))
            rmtree(model_path, ignore_errors=True)
            mlflow.sklearn.save_model(model, model_path)

            om = self.om
            # store with just the model path, specify the kind because paths can be other files too
            meta = om.models.put(model_path, 'mymodel', kind='mlflow.model')
            self.assertEqual(meta.kind, MLFlowModelBackend.KIND)
            model_ = om.models.get('mymodel')
            self.assertIsInstance(model_, mlflow.pyfunc.PyFuncModel)
            yhat_direct = model_.predict(reshaped(X))
            yhat_rt = om.runtime.model('mymodel').predict(X).get()
            assert_array_equal(yhat_rt, yhat_direct)

        def test_save_mlflow_saved_model_file(self):
            """ test deploying model saved by MLFlow, by file """
            import mlflow

            model_path = os.path.join(omegaml.defaults.OMEGA_TMP, 'mymodel')
            model = LinearRegression()
            X = pd.Series(range(0, 10))
            Y = pd.Series(X) * 2 + 3
            model.fit(reshaped(X), reshaped(Y))
            rmtree(model_path, ignore_errors=True)
            mlflow.sklearn.save_model(model, model_path)

            om = self.om
            # test multiple ways of storing
            for fn in ('mlflow://' + os.path.join(model_path, 'MLmodel'),
                       'mlflow://' + model_path):
                # store with just the MLmodel file as a reference, no kind necessary
                om.models.drop('mymodel', force=True)
                meta = om.models.put(fn, 'mymodel')
                self.assertEqual(meta.kind, MLFlowModelBackend.KIND)
                self.assertEqual(meta.kind, MLFlowModelBackend.KIND)
                model_ = om.models.get('mymodel')
                self.assertIsInstance(model_, mlflow.pyfunc.PyFuncModel)
                yhat_direct = model_.predict(reshaped(X))
                yhat_rt = om.runtime.model('mymodel').predict(X).get()
                assert_array_equal(yhat_rt, yhat_direct)

        def test_save_mlflow_pyfunc_model(self):
            """ test deploying a custom MLFlow PythonModel"""
            import mlflow

            class MyModel(mlflow.pyfunc.PythonModel):
                def predict(self, context, data):
                    return data

            X = pd.Series(range(10))
            model = MyModel()
            om = self.om
            meta = om.models.put(model, 'mymodel')
            self.assertEqual(meta.kind, MLFlowModelBackend.KIND)
            model_ = om.models.get('mymodel')
            self.assertIsInstance(model_, mlflow.pyfunc.PyFuncModel)
            yhat_direct = model_.predict(X)
            yhat_rt = om.runtime.model('mymodel').predict(X).get()
            assert_array_equal(yhat_rt, reshaped(yhat_direct))

        def test_inferred_model_flavor(self):
            """ test deploying an arbitrary model by inferring MLFlow flavor """
            import mlflow

            om = self.om
            model = LinearRegression()
            X = pd.Series(range(0, 10))
            Y = pd.Series(X) * 2 + 3
            model.fit(reshaped(X), reshaped(Y))
            meta = om.models.put(model, 'mymodel', kind='mlflow.model')
            self.assertEqual(meta.kind, MLFlowModelBackend.KIND)
            model_ = om.models.get('mymodel')
            self.assertIsInstance(model_, mlflow.pyfunc.PyFuncModel)

        def test_save_mlflow_model_run(self):
            """ test deploying an MLModel from a tracking server URI """
            import mlflow

            mlflow.set_tracking_uri(self.mlflow_tracking_db)
            mlflow.set_registry_uri(self.mlflow_registry_db)

            with mlflow.start_run() as run:
                model = LinearRegression()
                X = pd.Series(range(0, 10))
                Y = pd.Series(X) * 2 + 3
                model.fit(reshaped(X), reshaped(Y))
            mlflow.sklearn.log_model(sk_model=model,
                                     artifact_path='sklearn-model',
                                     registered_model_name='sklearn-model')

            # simulate a new session on another device (tracking URI comes from repo)
            mlflow.set_tracking_uri(None)

            om = self.om
            # use the tracking URI to store the model as a reference to a MLFlow tracking server
            meta = om.models.put('mlflow+models://sklearn-model/1', 'sklearn-model')
            self.assertEqual(meta.kind, MLFlowRegistryBackend.KIND)
            # simulate a new mlflow session
            model_ = om.models.get('sklearn-model')
            self.assertIsInstance(model_, mlflow.pyfunc.PyFuncModel)
            yhat_direct = model_.predict(reshaped(X))
            yhat_rt = om.runtime.model('sklearn-model').predict(X).get()
            assert_array_equal(yhat_rt, yhat_direct)

        def test_save_mlflow_model_run_file_tracking(self):
            """ test deploying an MLModel from a tracking URI using file path (not supported) """
            import mlflow

            om = self.om
            # note we don't set a tracking uri so mlflow uses a local file path and refused to
            # work with the file path as a model registry
            mlflow.set_tracking_uri(None)
            mlflow.set_registry_uri(None)
            with self.assertRaises(MlflowException):
                meta = om.models.put('mlflow+models://sklearn-model/1', 'sklearn-model')
            self.assertNotIn('sklearn-model', om.models.list())

        @classmethod
        def _clean_mlflowdbs(self):
            Path(self.mlflow_tracking_db.split('sqlite:///')[-1]).unlink(missing_ok=True)
            Path(self.mlflow_registry_db.split('sqlite:///')[-1]).unlink(missing_ok=True)

        @classmethod
        def _clean_mlruns(self):
            from mlflow.store import tracking
            mlruns_path = Path(__file__).parent / tracking.DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH
            shutil.rmtree(mlruns_path, ignore_errors=True)
            mlruns_path.mkdir(parents=True)

        def _enable_mlflow_exit_handling(self):
            # restore tracking uri to avoid mlflow exit handler error
            # https://github.com/mlflow/mlflow/issues/3755
            mlflow.set_tracking_uri(self.mlflow_tracking_db)
            mlflow.set_registry_uri(None)



from pymongo import MongoClient as RealMongoClient
from pymongo.errors import AutoReconnect, ConnectionFailure
from time import sleep
from urllib.parse import urlencode


def MongoClient(*args, **kwargs):
    """
    Shim function adding SSL kwargs on MongoClient instead of changing
    each and every location
    """
    from omegaml import settings
    defaults = settings()
    mongo_kwargs = dict(defaults.OMEGA_MONGO_SSL_KWARGS)
    mongo_kwargs.update(kwargs)
    return RealMongoClient(*args, **sanitize_mongo_kwargs(mongo_kwargs))


def sanitize_mongo_kwargs(kwargs):
    # keep kwargs sane
    # -- if we receive "use_ssl" = False from cloud config but have
    #    a local CA defined, remove it. Otherwise mongo raises ConfigurationError
    kwargs = dict(kwargs)
    if 'ssl' in kwargs:
        # ssl is alias to tls, should override tls (assume ssl is user-specified)
        # see https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html
        kwargs['tls'] = kwargs['ssl']
        del kwargs['ssl']
    if 'tlsCAFile' in kwargs and not kwargs.get('tls'):
        del kwargs['tlsCAFile']
    if 'ssl_ca_certs' in kwargs and not kwargs.get('tls'):
        del kwargs['ssl_ca_certs']
    return kwargs


def mongo_url(om):
    mongo_url = (om.datasets.mongo_url + '?authSource=admin&' +
                 urlencode(sanitize_mongo_kwargs(om.defaults.OMEGA_MONGO_SSL_KWARGS)))
    return mongo_url


def waitForConnection(client):
    _exc = None
    command = client.admin.command
    for i in range(100):
        try:
            # The ping command is cheap and does not require auth.
            command('ping')
        except (ConnectionFailure, AutoReconnect) as e:
            sleep(0.01)
            _exc = e
        else:
            _exc = None
            break
    if _exc is not None:
        raise _exc


